/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "vector_functions.h"

#include "InventorReader.h"
#include "Transformations/SceneNode.h"

static const unsigned addDummy = SceneLoader::instance()->addLoader(new InventorReader,"iv");

void InventorReader::InventorMaterial::reset() {
	ambient	 = make_float3(0.0f,0.0f,0.0f);
	diffuse  = make_float3(0.0f,0.0f,0.0f);
	specular = make_float3(0.0f,0.0f,0.0f);
	emmisive = make_float3(0.0f,0.0f,0.0f);
	shinines	 = 0.0f;
	reflectivity = 0.0f;
	transparency = 0.0f;
};

InventorReader::InventorReader(void) {
	tokenMap["Separator"]		   = LEX_SEPARATOR;

	tokenMap["ShapeHints"]		   = LEX_SHAPE_HINTS;
	tokenMap["shapeType"]		   = LEX_SHAPE_TYPE;
	tokenMap["UNKNOWN_SHAPE_TYPE"] = LEX_UNKNOWN_SHAPE_TYPE;
	tokenMap["faceType"]		   = LEX_FACE_TYPE;
	tokenMap["UNKNOWN_FACE_TYPE"]  = LEX_UNKNOWN_FACE_TYPE;

	tokenMap["Material"]	  = LEX_MATERIAL;
	tokenMap["ambientColor"]  = LEX_AMBIENT_COLOR;
	tokenMap["diffuseColor"]  = LEX_DIFFUSE_COLOR;
	tokenMap["emissiveColor"] = LEX_EMISSIVE_COLOR;
	tokenMap["specularColor"] = LEX_SPECULAR_COLOR;
	tokenMap["shininess"]	  =	LEX_SHININESS;
	tokenMap["transparency"]  =	LEX_TRANSPARENCY;
	

	tokenMap["IndexedFaceSet"] = LEX_INDEXED_MESH;
	tokenMap["vertexProperty"] = LEX_VERTEX_PROPERTY_VARIABLE;
	tokenMap["VertexProperty"] = LEX_VERTEX_PROPERTY_CLASS;
	tokenMap["vertex"]		   = LEX_VERTEX;

	tokenMap["normal"]				= LEX_NORMAL;
	tokenMap["normalBinding"]		= LEX_NORMAL_BINDING;
	tokenMap["materialBinding"]		= LEX_MATERIAL_BINDING;
	tokenMap["PER_VERTEX_INDEXED"]	= LEX_NORMAL_BINDING_PER_VERTEX_INDEXED;
	tokenMap["OVERALL"]				= LEX_MATERIAL_BINDING_OVERALL;

	tokenMap["coordIndex"]	= LEX_COORD_INDEX;
	tokenMap["normalIndex"]	= LEX_NORMAL_INDEX;
}

InventorReader::~InventorReader(void) {
}

void InventorReader::parseColor(float3 &dst) {
	nextToken(false); dst.x = redFloat;
	nextToken(false); dst.y = redFloat;
	nextToken(false); dst.z = redFloat;
};

void InventorReader::material() {
	currentMaterial.reset();

	while(true) {
		switch(token) {
			case LEX_AMBIENT_COLOR:  parseColor(currentMaterial.ambient);  nextToken(); break;
			case LEX_DIFFUSE_COLOR:  parseColor(currentMaterial.diffuse);  nextToken(); break;
			case LEX_EMISSIVE_COLOR: parseColor(currentMaterial.emmisive); nextToken(); break;
			case LEX_SPECULAR_COLOR: parseColor(currentMaterial.specular); nextToken(); break;
			case LEX_SHININESS:		 nextToken(false); currentMaterial.shinines		= redFloat; nextToken(); break;
			case LEX_TRANSPARENCY:	 nextToken(false); currentMaterial.transparency = redFloat; nextToken(); break;
			case LEX_CLOSE_CURLY: nextToken(); return;
		};
	};
};

void InventorReader::shapeHints() {
	while(true) {
		switch(token) {
			case LEX_SHAPE_TYPE:  nextToken(); nextToken(); break;
			case LEX_FACE_TYPE:   nextToken(); nextToken(); break;
			case LEX_CLOSE_CURLY: nextToken(); return;
		};
	};
};

void InventorReader::vertices() {
	GrowArray<float4>& verticesData = *verticesDataG;

	unsigned first  = verticesData.getNext();
	unsigned offset = first;

	while(true) {
		char *next = text + strcspn(text,",]");
		char  end  = *next;
		*next = 0;

		float4* dest = verticesData[offset];

		sscanf(text,"%f %f %f",&dest->x,&dest->y,&dest->z);
		dest->w = 1.0f;

		text = next+1;
		if(end == ']') break;

		offset = verticesData.getNext();
	};

	currentObject->vertexStart = first;
	currentObject->vertexCount = offset+1-first;

	nextToken();
};

void InventorReader::normals() {
	GrowArray<float4>& normalsData = *normalsDataG;

	unsigned first  = normalsData.getNext();
	unsigned offset = first;

	nextToken(false);

	if(token == LEX_OPEN_BRACKET) {
		while(true) {
			char *next = text + strcspn(text,",]");
			char  end  = *next;
			*next = 0;

			float4* dest = normalsData[offset];
			
			sscanf(text,"%f %f %f",&dest->x,&dest->y,&dest->z);
			dest->w = 0.0f;

			text = next+1;
			if(end == ']') break;

			offset = normalsData.getNext();
		};
		nextToken();
	} else {
		float4* dest = normalsData[offset];

		dest->x = redFloat; nextToken(false);
		dest->y = redFloat;	nextToken(false);
		dest->z = redFloat;	nextToken();
		dest->w = 0.0f;
	};

	currentObject->normalStart = first;
};

void InventorReader::geometry() {
	nextToken();
	nextToken();

	while(true) {
		switch(token) {
			case LEX_VERTEX: nextToken(); vertices(); break;
			case LEX_NORMAL: normals(); break;
			case LEX_NORMAL_BINDING: nextToken(); normalBinding(); break;
			case LEX_MATERIAL_BINDING: nextToken(); materialBinding(); break;
			case LEX_CLOSE_CURLY:  nextToken(); return;
		};
	};
};

void InventorReader::normalBinding() {
	nextToken();
};

void InventorReader::materialBinding() {
	nextToken();
};

void InventorReader::coordIndex() {
	GrowArray<unsigned>& coordData = *coordDataG;

	unsigned first  = coordData.getNext();
	unsigned face   = first;
	unsigned offset = coordData.getNext();
	unsigned faceCount = 0;

	unsigned count = 0;

	while(true) {
		char *next = text + strcspn(text,",]");
		char  end  = *next;
		*next = 0;

		int red = strtol(text,&text,10);
		count++;

		if(red == -1) {
			*coordData[face] = count-1;
			face  = offset;
			count = 0;
			faceCount++;
		} else {
			*coordData[offset] = red;
		};

		
		text = next+1;
		if(end == ']') break;
		
		offset = coordData.getNext();
	};

	*coordData[offset] = 0;
	currentObject->coorStart = first;
	currentObject->faceCount = faceCount;

	nextToken();
};

void InventorReader::normalIndex() {
	GrowArray<unsigned>& normData = *normDataG;

	unsigned first  = normData.getNext();
	unsigned face   = first;
	unsigned offset = normData.getNext();

	unsigned count = 0;

	nextToken();

	//if(token == LEX_OPEN_BRACKET) {
		while(true) {
			char *next = text + strcspn(text,",]");
			char  end  = *next;
			*next = 0;

			int red = strtol(text,&text,10);
			count++;
			if(red == -1) {
				*normData[face] = count-1;
				face  = offset;
				count = 0;
			} else {
				*normData[offset] = red;
			};

			text = next+1;
			if(end == ']') break;

			offset = normData.getNext();
		};
		nextToken();
	/*} else {
		int a = redInteger;
		nextToken(); a = redInteger;
		nextToken(); a = redInteger;
		nextToken();
	};*/
	*normData[offset] = 0;
	currentObject->normStart = first;
};

void InventorReader::mesh() {
	subObjects.push_back(SubObject());
	currentObject = &subObjects.back();
	currentObject->material = currentMaterial;

	while(true) {
		switch(token) {
			case LEX_VERTEX_PROPERTY_VARIABLE: nextToken(); geometry(); break;
			case LEX_COORD_INDEX:  nextToken(); coordIndex(); break;
			case LEX_NORMAL_INDEX: normalIndex(); break;
			case LEX_CLOSE_CURLY:  nextToken(); return;
		};
	};
};

void InventorReader::main() {
	while(true) {
		switch(token) {
			case LEX_SEPARATOR:    nextToken(); nextToken(); main(); break;
			case LEX_CLOSE_CURLY:  nextToken(); return;
			case LEX_SHAPE_HINTS:  nextToken(); nextToken(); shapeHints();	break; 
			case LEX_MATERIAL:	   nextToken(); nextToken(); material();	break; 
			case LEX_INDEXED_MESH: nextToken(); nextToken(); mesh();		break;
			case LEX_EOF:		   return;
		};
	};
};

struct Link {
	unsigned value;
	unsigned newVertex;
	Link*	 next;
};

struct Set {
	unsigned count;
	Link*	 values;

	unsigned insert(unsigned value,unsigned pos,unsigned& newVertex,TailPool<Link> &links);
};

unsigned Set::insert(unsigned value,unsigned pos,unsigned& newVertex,TailPool<Link> &links) {
	Link* link  = values;
	bool  found = false;

	while(link) {
		if(link->value == value) {
			found = true;
			break;
		};
		link=link->next;
	};

	if(!found) {
		Link* old = values;
		values = links.getMem();
		values->value     = value;
		values->newVertex = count ? newVertex++ : pos;
		values->next	  = old;

		count++;
	};

	return values->newVertex;
};

Object* InventorReader::makeObject() {
	GrowArray<unsigned>& coordData  = *coordDataG;
	GrowArray<unsigned>& normData   = *normDataG;
	GrowArray<float4>& verticesData = *verticesDataG;
	GrowArray<float4>& normalsData  = *normalsDataG;

	TailPool<Link> links(16*1024);

	unsigned vertexCount = 0;
	unsigned faceCount   = 0;
	for(size_t i = 0;i<subObjects.size();i++) {
		SubObject& current  = subObjects[i];

		Set* set = new Set[current.vertexCount];
		memset(set,0,sizeof(Set)*current.vertexCount);

		unsigned* indices = coordData[current.coorStart];
		unsigned* normals = normData [current.normStart];

		unsigned newVertex = current.vertexCount;

		while(*indices) {
			if(*indices++ != 3) throw 1;
			if(*normals++ != 3) throw 1;

			unsigned face,normal;
			face = *indices++; normal = *normals++; set[face].insert(normal,face,newVertex,links);
			face = *indices++; normal = *normals++; set[face].insert(normal,face,newVertex,links);
			face = *indices++; normal = *normals++; set[face].insert(normal,face,newVertex,links);
		};

		vertexCount += newVertex;
		faceCount   += current.faceCount;

		delete[] set;
		links.reset();
	};

	Object *object		= new Object;
	object->vertices	= new Vertex[vertexCount];
	object->normal		= new Vertex[vertexCount];
	object->faces		= new Triangle[faceCount];
	object->vertexCount = vertexCount;
	object->facesCount	= faceCount;
	//object->materials.push_back(0);

	vertexCount = 0;
	faceCount	= 0;

	unsigned objectMaterial = 0;
	std::vector<Material> &globalMaterials = MaterialManager::instance()->getMaterials();

	Vertex* currentVertex = object->vertices;
	for(size_t i = 0;i<subObjects.size();i++) {
		SubObject& current  = subObjects[i];

		Set* set = new Set[current.vertexCount];
		memset(set,0,sizeof(Set)*current.vertexCount);

		unsigned* indices = coordData[current.coorStart];
		unsigned* normals = normData [current.normStart];

		unsigned newVertex = current.vertexCount;

		while(*indices) {
			if(*indices++ != 3) throw 1;
			if(*normals++ != 3) throw 1;

			unsigned face,normal;
			face = *indices; normal = *normals++; *indices++ = set[face].insert(normal,face,newVertex,links);
			face = *indices; normal = *normals++; *indices++ = set[face].insert(normal,face,newVertex,links);
			face = *indices; normal = *normals++; *indices++ = set[face].insert(normal,face,newVertex,links);
		};

		float4* vertices	 = verticesData[current.vertexStart];
		float4* normalVector = normalsData [current.normalStart];

		for(unsigned j=0;j<current.vertexCount;j++) {
			Link* link  = set[j].values;
			while(link) {
				object->vertices[vertexCount+link->newVertex].x = vertices[j].x;
				object->vertices[vertexCount+link->newVertex].y = vertices[j].y;
				object->vertices[vertexCount+link->newVertex].z = vertices[j].z;
				object->vertices[vertexCount+link->newVertex].w = vertices[j].w;
				object->normal  [vertexCount+link->newVertex].x = normalVector[link->value].x;
				object->normal  [vertexCount+link->newVertex].y = normalVector[link->value].y;
				object->normal  [vertexCount+link->newVertex].z = normalVector[link->value].z;
				object->normal  [vertexCount+link->newVertex].w = normalVector[link->value].w;

				link=link->next;
			};
		};

		indices = coordData[current.coorStart];
		for(unsigned j=0;j<current.faceCount;j++) {
			indices++;
			object->faces[faceCount+j].a = *indices++ + vertexCount;
			object->faces[faceCount+j].b = *indices++ + vertexCount;
			object->faces[faceCount+j].c = *indices++ + vertexCount;
			object->faces[faceCount+j].material = objectMaterial;
		};

		globalMaterials.push_back(Material());
		object->materials.push_back(globalMaterials.size()-1);
		Material& material = globalMaterials.back();
		material.ambient	  = current.material.ambient;
		material.diffuse	  = current.material.diffuse;
		material.specular	  = current.material.specular;
		material.shinines	  = current.material.shinines;
		material.reflectivity = 0.0f; 
		if(current.material.emmisive.x!=0 || current.material.emmisive.y!=0 || current.material.emmisive.z!=0)
			material.diffuse = current.material.emmisive;

		float specular = (material.specular.x + material.specular.y + material.specular.z)/3.0f;
		float specularMax = max(max(material.specular.x,material.specular.y),material.specular.z);
		if(specularMax>0.30f)
			material.reflectivity = specular;


		vertexCount += newVertex;
		faceCount   += current.faceCount;
		objectMaterial++;

		delete[] set;
		links.reset();
	};

	__m128 minimum,maximum;
	minimum = minimumInit3;
	maximum = maximumInit3;

	for(unsigned i=0;i<object->vertexCount;i++) {
		__m128 current = sse::load<__m128>(&object->vertices[i]);
		minimum = sse::min(minimum,current);
		maximum = sse::max(maximum,current);
	};

	object->minimum.vec = minimum;
	object->maximum.vec = maximum;

	return object;
};

Object* InventorReader::loadFile(const char *fileName) {
	FILE*  input;
	size_t fileSize;

	input = fopen(fileName,"r");

	fseek(input,0,SEEK_END);
	fileSize = ftell(input);
	fseek(input,0,SEEK_SET);

	char* buffer = new char[fileSize+1];

	fileSize = fread(buffer,1,fileSize,input);
	fclose(input);

	buffer[fileSize]=0;


	verticesDataG = new GrowArray<float4>(32768);
	normalsDataG  = new GrowArray<float4>(32768);
	coordDataG	  = new GrowArray<unsigned>(65536);
	normDataG	  = new GrowArray<unsigned>(65536);

	text = buffer;
	nextToken();
	main();

	delete buffer;

	Object *object = makeObject();

	delete verticesDataG;
	delete normalsDataG;
	delete coordDataG;
	delete normDataG;
	subObjects.resize(0);

	return object;
};