/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "TriLoader.h"
#include "Geometry.h"

#include <algorithm>
#include "Utils/iterator.h"
#include "Utils/memory.h"
#include <math.h>

#include <time.h>

#include "Transformations/SceneNode.h"

static const unsigned addDummy = SceneLoader::instance()->addLoader(new TriLoader,"tri2");

TriLoader::TriLoader(void) {
}

TriLoader::~TriLoader(void){
}

class VertexIndex:public SPVector,public AllignedS {
public:
	unsigned face;
	unsigned vertex;
	unsigned start;
	unsigned count;

	bool operator<(const VertexIndex &right) {
		if(x < right.x) return true;
		if(x > right.x) return false;
		if(y < right.y) return true;
		if(y > right.y) return false;
		if(z < right.z) return true;
		return false;
	};

	bool operator==(const VertexIndex &right) {
		return x==right.x && y==right.y && z==right.z;
	};
};

struct FaceIndex {
	unsigned face;
	unsigned index;
};

struct Link {
	unsigned  face;
	unsigned  index;
	Link	 *next;
};

struct ShadeGroup {
	Link*	first;
	Link*	last;
};

bool connected(const Triangle &left,const Triangle& right) {
	unsigned common=0;
	common+= left.a == right.a;
	common+= left.a == right.b;
	common+= left.a == right.c;
	common+= left.b == right.a;
	common+= left.b == right.b;
	common+= left.b == right.c;
	common+= left.c == right.a;
	common+= left.c == right.b;
	common+= left.c == right.c;
	return common==2;
};	

bool join(ShadeGroup &left,ShadeGroup &right,Triangle *faces,Vertex* normals) {
	if(connected(faces[left.last->face],faces[right.first->face])) {
		float angle = normals[left.last->face] * normals[right.first->face];
		if(fabs(angle) > 0.5f) {
			left.last->next = right.first;
			left.last		= right.last;
			return true;
		} else {
			return false;
		};
	};
	if(connected(faces[left.last->face],faces[right.last->face])) {
		float angle = normals[left.last->face] * normals[right.last->face];
		if(fabs(angle) > 0.5f) {
			Link* last = 0; 
			Link* cur  = right.first;
			while(cur) {
				Link* newLast = cur->next;
				cur->next = last;
				last = cur;
				cur = newLast;
			};

			left.last->next = right.last;
			left.last		= right.first;
			return true;
		} else {
			return false;
		};
	};
	return false;
};

Object* TriLoader::loadFile(const char *fileName) {
	clock_t start,stop;

	FILE*  input;
	size_t fileSize;

	input = fopen(fileName,"r");

	fseek(input,0,SEEK_END);
	fileSize = ftell(input);
	fseek(input,0,SEEK_SET);

	char* buffer = new char[fileSize+1];
	char* cur    = buffer;

	fileSize = fread(buffer,1,fileSize,input);
	fclose(input);

	buffer[fileSize]=0;

	unsigned facesCount;
	unsigned vertexCount;

	sscanf(buffer,"%i TRIANGLES",&facesCount);
	cur = strchr(cur,'\n')+1;

	vertexCount = facesCount*3;

	__m128 minimum,maximum;
	minimum = minimumInit3;
	maximum = maximumInit3;

	Object *object		= new Object;
	object->faces		= new Triangle[facesCount];
	object->facesCount	= facesCount;
	object->materials.push_back(0);

	Triangle* __restrict faces		= new Triangle[facesCount];
	VertexIndex*		 vertices	= new VertexIndex[vertexCount];

	start = clock();

	for(unsigned i=0,j=0;i<facesCount;i++,j+=3) {
		char* next = strchr(cur,'\n'); 
		*next = 0;

		sscanf(cur,"TRIANGLE (%f,%f,%f)(%f,%f,%f)(%f,%f,%f)",
			&vertices[j  ].x,&vertices[j  ].y,&vertices[j  ].z,
			&vertices[j+1].x,&vertices[j+1].y,&vertices[j+1].z,
			&vertices[j+2].x,&vertices[j+2].y,&vertices[j+2].z);
		cur = next+1; 
		vertices[j  ].face   = i;
		vertices[j  ].vertex = 0;
		vertices[j  ].w		 = 1.0f;
		vertices[j+1].face   = i;
		vertices[j+1].vertex = 1;
		vertices[j+1].w		 = 1.0f;
		vertices[j+2].face   = i;
		vertices[j+2].vertex = 2;
		vertices[j+2].w		 = 1.0f;

		faces[i].a = j;
		faces[i].b = j+1;
		faces[i].c = j+2;
		object->faces[i].material = 0;
		
		//Update bounding box coordinates
		__m128 tri1 = _mm_load_ps(&vertices[j  ].x);
		__m128 tri2 = _mm_load_ps(&vertices[j+1].x);
		__m128 tri3 = _mm_load_ps(&vertices[j+2].x);
		minimum = _mm_min_ps(minimum,tri1);
		maximum = _mm_max_ps(maximum,tri1);
		minimum = _mm_min_ps(minimum,tri2);
		maximum = _mm_max_ps(maximum,tri2);
		minimum = _mm_min_ps(minimum,tri3);
		maximum = _mm_max_ps(maximum,tri3);
	};

	stop = clock();
	printf("parsing: %f\n",(double)(stop-start)/CLOCKS_PER_SEC);

	start = clock();

	PlainIterator<VertexIndex>beg(vertices);
	PlainIterator<VertexIndex>end(vertices+vertexCount);
	std::sort(beg,end);

	stop = clock();
	printf("sorting: %f\n",(double)(stop-start)/CLOCKS_PER_SEC);

	FaceIndex* faceIndex  = new FaceIndex[vertexCount];

	unsigned reduceCount = 0;
	unsigned thisCount	 = 0;

	start = clock();

	vertices[0].start = 0;
	for(unsigned i=0;i<vertexCount;i++) {
		if(i==vertexCount-10)
			i=i;

		if(!(vertices[i-1] == vertices[i]) && i!=0) {
			vertices[reduceCount].count = thisCount;
			thisCount = 0;

			reduceCount++;
			vertices[reduceCount]=vertices[i];
			vertices[reduceCount].start=i;
		};

		faceIndex[i].face  = vertices[i].face;
		faceIndex[i].index = vertices[i].vertex;
		thisCount++;
		
		switch(vertices[i].vertex) {
			case 0: faces[vertices[i].face].a = reduceCount; break;
			case 1: faces[vertices[i].face].b = reduceCount; break;
			case 2: faces[vertices[i].face].c = reduceCount; break;
		};
	};
	vertices[reduceCount++].count = thisCount;


	stop = clock();
	printf("merging: %f\n",(double)(stop-start)/CLOCKS_PER_SEC);

	start = clock();

	Vertex* normals = new Vertex[facesCount];
	for(unsigned i=0;i<facesCount;i++) {
		unsigned v1,v2,v3;
		v1 = faces[i].a;
		v2 = faces[i].b;
		v3 = faces[i].c;

		SPVector a = vertices[v1].vec;
		SPVector b = vertices[v2].vec;
		SPVector c = vertices[v3].vec;

		b-=a;
		c-=a;

		SPVector normal = b % c;
		normal.Normalize();
		normals[i].vec=normal.vec;
	};

	stop = clock();
	printf("normals: %f\n",(double)(stop-start)/CLOCKS_PER_SEC);

	object->vertices  = new Vertex[vertexCount];
	unsigned newCount = 0;

	Link		   links[128];
	ShadeGroup	   shadeGroups[128];

	start = clock();

	for(unsigned i=0;i<reduceCount;i++) {
		FaceIndex* currentVertex = &faceIndex[vertices[i].start];
		unsigned   count		 =  vertices[i].count;
		unsigned   groupsCount   =  count;

		for(unsigned j=0;j<count;j++) {
			shadeGroups[j].first = shadeGroups[j].last = &links[j];
			links[j].face  = currentVertex[j].face;
			links[j].index = currentVertex[j].index;
			links[j].next  = 0;

			if(currentVertex[j].face == 792)
				j=j;
		};

		bool changed = false;

		while(true) {
			for(unsigned j=0;j<groupsCount;j++) {
				for(unsigned k=0;k<groupsCount && j<groupsCount;k++) {
					if(j!=k) {
						if(join(shadeGroups[j],shadeGroups[k],faces,normals)) {
							groupsCount--;
							if(k!=groupsCount) {
								shadeGroups[k]=shadeGroups[groupsCount];
								k--;
							};
							changed = true;
						};
					};
				};
			};
			if(!changed) break;
			changed = false;
		};

		for(unsigned j=0;j<groupsCount;j++) {
			Link *current = shadeGroups[j].first;

			object->vertices[newCount].vec = vertices[i].vec;
			do {
				switch(current->index) {
					case 0: object->faces[current->face].a = newCount; break;
					case 1: object->faces[current->face].b = newCount; break;
					case 2: object->faces[current->face].c = newCount; break;
				};
				current = current->next;
			} while (current);
			newCount++;
		};
	};

	stop = clock();
	printf("shade groups: %f\n",(double)(stop-start)/CLOCKS_PER_SEC);


	object->vertexCount = newCount;
	object->normal		= new Vertex[newCount];

	for(unsigned i=0;i<facesCount;i++) {
		unsigned v1,v2,v3;
		v1 = object->faces[i].a;
		v2 = object->faces[i].b;
		v3 = object->faces[i].c;

		object->normal[v1] += normals[i];
		object->normal[v2] += normals[i];
		object->normal[v3] += normals[i];
	};

	for(unsigned i=0;i<newCount;i++)
		object->normal[i].Normalize();

	object->minimum.vec = minimum;
	object->maximum.vec = maximum;

	return object;
};
