/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>

#include "Wrapper.h"

#include "SceneLights.h"

#include "Material.h"
#include "CameraCuda.h"

#include <vector_types.h>
#include <cuda_runtime_api.h>

#define GRID
#ifdef GRID

//These are definitions of host functions in cuda source code. Probably should be moved to some header file but...
float cudaRenderGrid(unsigned int width,unsigned int height,
					  float3 &minimum, float3 &maximum, float cellSize, uint3 &dim,
					  cudaArray* cudaGrid,uint4* triangles,cudaArray* vertices,cudaArray *normals,
					  CameraCuda &camera,unsigned lightCount,Light* lightsIn,Material* materials,
					  unsigned int *buffer,unsigned int *copy,RenderMode mode);

float cudaRenderGridTexture(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum, float cellSize, uint3 &dim,
				    cudaArray* cudaGrid,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode);

CudaRenderer::CudaRenderer(unsigned width,unsigned height,int pbo):width(width),height(height),pbo(pbo) {
	renderTarget = pbo==-1 ? MEMORY : PBO;

	//If render target is memory, allocate the buffer.
	if(renderTarget==MEMORY) {
		cudaMallocHost((void**)&copy    ,width*height*4);	//Use pinned memory for faster bud data transfer.
		cudaMalloc	  ((void**)&buffer  ,width*height*4);
	};
};

void CudaRenderer::resize(unsigned newWidth,unsigned newHeight) {
	width  = newWidth;
	height = newHeight;

	//If current render target is memory, free the old buffer and allocated new one.
	if(renderTarget == MEMORY) {
		cudaFreeHost(copy);
		cudaFree	(buffer);

		cudaMallocHost((void**)&copy    ,width*height*4);
		cudaMalloc	  ((void**)&buffer  ,width*height*4);
	};
};

void CudaRenderer::setRenderTarget(unsigned newPbo) {
	if(renderTarget == MEMORY && newPbo!=-1) {
		cudaFreeHost(copy);
		cudaFree	(buffer);
	};

	pbo = newPbo;
	renderTarget = newPbo==-1 ? MEMORY : PBO;

	if(renderTarget==MEMORY) {
		cudaMallocHost((void**)&copy    ,width*height*4);
		cudaMalloc	  ((void**)&buffer  ,width*height*4);
	} else 
		copy = 0;
};
	
CudaRenderer::~CudaRenderer() {
	if(renderTarget == MEMORY) {
		cudaFreeHost(copy);
		cudaFree	(buffer);
	};
};


CudaRenderGrid::CudaRenderGrid(unsigned width,unsigned height,UniformGrid *grid,int pbo):CudaRenderer(width,height,pbo),grid(grid) {
	grid->prepareCudaTexture(minimum,maximum,cellSize,dim,cudaGrid,triangles,vertices,normals);
	materials = MaterialManager::instance()->copyToCuda();
};

CudaRenderGrid::~CudaRenderGrid() {
	MaterialManager::instance()->clearCuda();
	delete grid;
};

float CudaRenderGrid::render(Camera& camera,SceneLights &lights,RenderMode mode) {
	unsigned lightCount = (unsigned)lights.size();
	Light *lightBegin   = &lights.front();
	
	CameraCuda cameraCuda;
	camera.prepareCuda(cameraCuda);

	if(renderTarget == MEMORY)
		return cudaRenderGrid(width,height,minimum,maximum,cellSize,dim,cudaGrid,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,buffer,copy,mode);
	else
		return cudaRenderGridTexture(width,height,minimum,maximum,cellSize,dim,cudaGrid,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,pbo,mode);
};

#endif

#define KDTREE
#ifdef KDTREE
float cudaRenderKd(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum,
				    cudaArray* cudaTree,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int *buffer,unsigned int *copy,RenderMode mode);

float cudaRenderKdTexture(unsigned int width,unsigned int height,
				    float3 &minimum, float3 &maximum,
				    cudaArray* cudaTree,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode);

CudaRenderKd::CudaRenderKd(unsigned width,unsigned height,KdTree *tree,int pbo):CudaRenderer(width,height,pbo),tree(tree) {
	tree->prepareCudaTexture(minimum,maximum,cudaTree,triangles,vertices,normals);
	materials = MaterialManager::instance()->copyToCuda();
};

CudaRenderKd::~CudaRenderKd() {
	delete tree;
	MaterialManager::instance()->clearCuda();
};

float CudaRenderKd::render(Camera& camera,SceneLights &lights,RenderMode mode) {
	unsigned lightCount = (unsigned)lights.size();
	Light *lightBegin   = &lights.front();
	
	CameraCuda cameraCuda;
	camera.prepareCuda(cameraCuda);

	if(renderTarget == MEMORY)
		return cudaRenderKd(width,height,minimum,maximum,cudaTree,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,buffer,copy,mode);
	else 
		return cudaRenderKdTexture(width,height,minimum,maximum,cudaTree,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,pbo,mode);
};

#endif

#define BVHTREE
#ifdef  BVHTREE
float cudaRenderBvh(unsigned int width,unsigned int height,
				    unsigned* cells,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int *buffer,unsigned int *copy,RenderMode mode);
float cudaRenderBvhTexture(unsigned int width,unsigned int height,
				    unsigned* cells,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode);


CudaRenderBvh::CudaRenderBvh(unsigned width,unsigned height,Bvh *tree,int pbo):CudaRenderer(width,height,pbo),tree(tree) {
	tree->prepareCuda(minimum,maximum,bvhTree,triangles,vertices,normals);
	materials = MaterialManager::instance()->copyToCuda();
};

CudaRenderBvh::~CudaRenderBvh() {
	delete tree;
	MaterialManager::instance()->clearCuda();
};

float CudaRenderBvh::render(Camera& camera,SceneLights &lights,RenderMode mode) {
	unsigned lightCount = (unsigned)lights.size();
	Light *lightBegin   = &lights.front();
	
	CameraCuda cameraCuda;
	camera.prepareCuda(cameraCuda);

	if(renderTarget == MEMORY)
		return cudaRenderBvh(width,height,bvhTree,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,buffer,copy,mode);
	else 
		return cudaRenderBvhTexture(width,height,bvhTree,triangles,vertices,normals,cameraCuda,lightCount,lightBegin,materials,pbo,mode);
};

#endif