/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <float.h>

struct BvhCell {
	float3	 minimum;
	float3   maximum;
	unsigned children;
	unsigned count;
};

__device__
void reduction(int* data) {
	int thid  = threadIdx.y*8 + threadIdx.x;
	__syncthreads();
#ifdef __DEVICE_EMULATION__
	if(thid == 0) {
		unsigned sum=0;
		for(int i=0;i<64;i++)
			sum+=data[i];
		data[0]=sum;
	};
	__syncthreads();
#else
	if(thid<32) {
		data[thid] += data[thid + 32];
		data[thid] += data[thid + 16];
		data[thid] += data[thid +  8];
		data[thid] += data[thid +  4];
		data[thid] += data[thid +  2];
		data[thid] += data[thid +  1]; 
	};
	__syncthreads();
#endif
}

__device__
void intersectDirect(int a,int b,int c,float3 &eye,float3 &dir,float &u,float &v,float &t) {
	unsigned x1 = a &  255;
	unsigned y1 = a >> 8;
	unsigned x2 = b &  255;
	unsigned y2 = b >> 8;
	unsigned x3 = c &  255;
	unsigned y3 = c >> 8;

	float3 v0 = float4to3(tex2D(vertexTexture,x1,y1));
	float3 v1 = float4to3(tex2D(vertexTexture,x2,y2));
	float3 v2 = float4to3(tex2D(vertexTexture,x3,y3));

	v2 -= v0;
	v1 -= v0;

	float3 tvec   = eye - v0;
	float3 pvec   = cross(dir,v2);
	float  invDet = 1.0f / dot(v1,pvec);

	u = dot(tvec,pvec)*invDet;

	float3 qvec	  = cross(tvec,v1);
	t = dot(v2,qvec)*invDet;
	v = dot(dir,qvec)*invDet;
};

__device__ void loadBvh(BvhCell *dst,unsigned *cells,unsigned node,unsigned index) {
	unsigned start = node & 0xFFFFFFFE;
	unsigned data  = cells[start*8+index];

	if( (node&1) && index >= 8)
		((unsigned*)dst)[index - 8] = data;
	if(!(node&1) && index <  8)
		((unsigned*)dst)[index]	 = data;

};

#ifdef __DEVICE_EMULATION__
	#define EMU_SYNC __syncthreads();
#else
	#define EMU_SYNC 
#endif

template<bool output>
__device__ int traverseBvh(float3 &eye,float3 &dir,unsigned *cells, uint4* triangles, unsigned* stackNode, unsigned* stackCount,bool idle) {
	__shared__ BvhCell  children[2];
	__shared__ int	    resolve[64];
	__shared__ int	    resolveNext[4];
	__shared__ unsigned stackPos;			//This could be stored per thread. In that case some synchronizations and branching
											//would be eliminated at the cost of one more registers. Performance wise probably unimportant.

	unsigned current;

	int      hitTriangle = -1;
	int		 index = (threadIdx.y*8 + threadIdx.x);
	float    closestHit = FLT_MAX;
	
	if(index==0) stackPos=0;

	if(index < 16)
		loadBvh(&children[0],cells,0,index);

	__syncthreads();

	unsigned offspring = children[0].children;
	unsigned count     = children[0].count;

	while(true) {
	 	if(count > 0) {	//If the primitive count is non zero the current node is leaf.
			if(index < 16) {
				//The maximum number of triangles in node is 4 so the size of indexes is 64 bytes. Read them with one coallesced access.
				int data = ((int*)triangles)[offspring*4+index];
				resolve[index]=data;
			};

			__syncthreads();

			//For all the trinagles try to find the intersection.
			for(int i=0;i<count;i++) {
				float u,v,t;
				int a = resolve[i*4+1];
				int b = resolve[i*4+2];
				int c = resolve[i*4+3];
			
				intersectDirect(a,b,c,eye,dir,u,v,t);

				if(v>=-0.00048828125f && u>=-0.00048828125f && u+v<=1.0000304f && t>=0.010070801f && t<=closestHit) {
					closestHit  = t;
					hitTriangle = offspring; 
				};
				
				offspring++;
			};

			EMU_SYNC

			if(stackPos>0) {
				offspring = stackNode [stackPos-1];
				count	  = stackCount[stackPos-1];

				//All threads must read correct item from stack, so we must synchronize before decreasing it.
				__syncthreads();	
				if(index==0) stackPos--;
				__syncthreads();
			} else 
				break;
		} else {
			EMU_SYNC

			//Internal node
			if(index < 16) {
				unsigned data  = cells[offspring*8+index];
				((unsigned*)children)[index] = data;
			};

			__syncthreads();

			float leftMin ,leftMax;
			float rightMin,rightMax;

			intersectBox(children[0].minimum,children[0].maximum,eye,dir,leftMin ,leftMax );
			intersectBox(children[1].minimum,children[1].maximum,eye,dir,rightMin,rightMax);

			int wantLeft  = (leftMin  <= leftMax ) && (leftMin  < closestHit) && (leftMax >=0) && idle;
			int wantRight = (rightMin <= rightMax) && (rightMin < closestHit) && (rightMax>=0) && idle;

			if(index<4) resolveNext[index]=0;
			__syncthreads();
			resolveNext[2*wantLeft + wantRight]=1;
			__syncthreads();

			if(resolveNext[3] || (resolveNext[1] && resolveNext[2])) {
				resolve[index] = (wantRight && (rightMin<leftMin  || !wantLeft  )) - 
							     (wantLeft  && (leftMin <rightMin || !wantRight ));
				if(!idle) resolve[index]=0;
				
				reduction(resolve);

				unsigned pushNode;
				unsigned pushCount;
				if(resolve[0]<0) {
					count		= children[0].count;
					offspring	= children[0].children;
					pushCount   = children[1].count;
					pushNode	= children[1].children;
					
				} else {
					count		= children[1].count;
					offspring	= children[1].children;
					pushCount   = children[0].count;
					pushNode	= children[0].children;
				};

				if(index==0) {
					stackNode [stackPos  ] = pushNode;
					stackCount[stackPos++] = pushCount;
				};

				__syncthreads();
			}
			else if(resolveNext[1]) {offspring = children[1].children; count = children[1].count;}
			else if(resolveNext[2]) {offspring = children[0].children; count = children[0].count;}
			else {
				if(stackPos>0) {
					offspring = stackNode [stackPos-1];
					count	  = stackCount[stackPos-1];

					__syncthreads();
					if(index==0) stackPos--;
					__syncthreads();
				} else 
					break;
			};
		};
		EMU_SYNC
	};
	EMU_SYNC

	if(output)
		return idle ? hitTriangle : -1;
	else
		return hitTriangle == -1 ? __float_as_int(-1.0f) :  __float_as_int(closestHit);
};

__device__ void catchNormalDirect(unsigned a,unsigned b,unsigned c,float u,float v,float3& normal) {
	//Normals are stored in 2D texture(Due to the 1D size limitation) with width of 256. Convert linear indexes to 2D coordinates.
	unsigned x1 = a &  255;
	unsigned y1 = a >> 8;
	unsigned x2 = b &  255;
	unsigned y2 = b >> 8;
	unsigned x3 = c &  255;
	unsigned y3 = c >> 8;

	//Read normals from texture
	float3 normal0 = float4to3(tex2D(normalTexture,x1,y1));
	float3 normal1 = float4to3(tex2D(normalTexture,x2,y2));
	float3 normal2 = float4to3(tex2D(normalTexture,x3,y3));

	//The intesection point is given is barycentric coordinates and is therefore weighted average of vertices.
	//The third missing coordinate can be computed from the supplied two as sum must be 1.
	float w = 1.0f - u - v;
	normal0*=w;
	normal1*=u;
	normal2*=v;

	//sum normals and renormalize
	normal0+=normal1;
	normal0+=normal2;
	normalize(normal0);

	normal = normal0;
};

__global__ void renderBvhP( float3 eye	   ,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
						    unsigned* cells,uint4* triangles,Material* materials,
						    unsigned int* buffer) {
	__shared__ unsigned	stackNode [128];
	__shared__ unsigned	stackCount[128];

	unsigned material;
    float3 dir;
	float3 origin = eye;
	
	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;
	
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	//Traverse the grid and try to find intersection.
	int hitTriangle = traverseBvh<true>(origin,dir,cells,triangles,stackNode,stackCount,true);

	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	if(hitTriangle >= 0) {
		float  t;
		float3 normal;

		//Get intersection parameters for hit triangle. This could be suplied from "traverseTexture" 
		//but is done this way to save some registers. Could be changed to descibed method lates as we 
		//still have some free registers left to next limit and are far away from lowering register usage
		//to get to previous limit.
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		float3 intersectionPoint = origin + dir*t;
		float3 reflectedDir = dir - normal*dot(normal,dir)*2.0f;		//Reflected ray computation

		for(int i=0;i<lightsCount;i++) {
			dir = intersectionPoint - lights[i].possition;
			shadePrimary(color,i,materials,material,normal,dir,reflectedDir);
		};
	};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

__global__ void renderBvhPS(float3 eye	   ,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
						     unsigned* cells,uint4* triangles,Material* materials,
						     unsigned int* buffer) {
	__shared__ unsigned	stackNode [128];
	__shared__ unsigned	stackCount[128];

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;
	
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	//if(!(blockIdx.x == 22 && blockIdx.y == 22))	return;
	//if(blockIdx.x < 10 || blockIdx.x > 30 || blockIdx.y <20 || blockIdx.y>35) return;

	bool lastHit = true;

	int hitTriangle = traverseBvh<true>(eye,dir,cells,triangles,stackNode,stackCount,true);

	unsigned material;
	float3   intersectionPoint;
	float3	 reflectedDir;
	float3   normal;

	if(hitTriangle >=0) {
		float  t;
		
		catchNormalIntersection(triangles,hitTriangle,eye,dir,normal,t,material);

		intersectionPoint = eye + dir*t;						//Compute intersection point of ray and triangle.
		reflectedDir	  = dir - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
	};

	unsigned shadowPass = 0;
	float3   color		= make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	short index  = threadIdx.y*8 + threadIdx.x;
	bool  rayHit = hitTriangle >= 0;
	__shared__ int hit;
	if(index == 0) hit = 0;
	__syncthreads();
	if(rayHit)	   hit = 1;
	__syncthreads();

	if(hit)
		while(shadowPass < lightsCount) {
			float3 origin = lights[shadowPass].possition;
 			float3 dir	  = intersectionPoint - origin;

			float t = __int_as_float(traverseBvh<false>(origin,dir,cells,triangles,stackNode,stackCount,rayHit));
	
			if(rayHit) {
				bool shadow = t>=0.0f && t<0.999f;
				shade(color,shadow,shadowPass,materials,material,1.0f,1.0f,normal,dir,reflectedDir);
			};

			shadowPass++;
		};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

__global__ void renderBvhPSR(float3 eye    ,float3 upperLeft, float3 dx, float3 dy,unsigned short startX,unsigned short startY,unsigned width,
						     unsigned* cells,uint4* triangles,Material* materials,
						     unsigned int* buffer) {
	__shared__ unsigned	stackNode [128];
	__shared__ unsigned	stackCount[128];

	unsigned	reflectLimit = 2;		//Limit number of reflection to fixed number. Might change to paramtere later
	float		refletSum	 = 1.0f;	//

	//Compute pixel coordinates. The grid is composed of 8x8 blocks.
	unsigned short x  = startX + blockIdx.x*8 + threadIdx.x;
	unsigned short y  = startY + blockIdx.y*8 + threadIdx.y;
	
	//Compute ray direction for current pixel. To achive right traversal, use higher precision intrinsic.
	float3 dir;
	dir.x = upperLeft.x + __fmul_rn(dx.x,x) + __fmul_rn(dy.x,y);
	dir.y = upperLeft.y + __fmul_rn(dx.y,x) + __fmul_rn(dy.y,y);
	dir.z = upperLeft.z + __fmul_rn(dx.z,x) + __fmul_rn(dy.z,y);

	float3 color = make_float3(0.0f,0.0f,0.0f);	//Start with black pixel color

	//if(!(blockIdx.x == 31 && blockIdx.y == 24))	return;

	bool  rayHit  = true;
	float3 origin = eye;
	while(true) {
		int hitTriangle = traverseBvh<true>(origin,dir,cells,triangles,stackNode,stackCount,rayHit);

		unsigned material;
		float3   intersectionPoint;
		float3	 reflectedDir;
		float3   normal;

		rayHit = hitTriangle >= 0;

		if(rayHit) {
			float  t;
			
			catchNormalIntersection(triangles,hitTriangle,origin,dir,normal,t,material);

			intersectionPoint = origin + dir*t;							//Compute intersection point of ray and triangle.
			reflectedDir	  = dir    - normal*dot(normal,dir)*2.0f;	//Reflected ray computation
		};

		unsigned shadowPass = 0;

		short index  = threadIdx.y*8 + threadIdx.x;
		__shared__ int hit;
		if(index == 0) hit = 0;
		__syncthreads();
		if(rayHit)	   hit = 1;
		__syncthreads();

		if(hit) {
			while(shadowPass < lightsCount) {
				origin = lights[shadowPass].possition;
				dir    = intersectionPoint - origin;

				float t = __int_as_float(traverseBvh<false>(origin,dir,cells,triangles,stackNode,stackCount,rayHit));
		
				if(rayHit) {
					bool shadow = t>=0.0f && t<0.999f;
					float currentVal = refletSum*(1.0f - materials[material].reflectivity);
					shade(color,shadow,shadowPass,materials,material,currentVal,refletSum,normal,dir,reflectedDir);
				};

				shadowPass++;
			};

			float reflectivity=0.0f;
			if(rayHit) reflectivity=materials[material].reflectivity;

			//We have looped over all the lights. Prepare for reflected ray if the surface is reflective.
			//Exit if number of reflection limit was reached or the surface is not reflective.
			rayHit = rayHit && reflectLimit != 0 && reflectivity != 0.0f;

			if(index == 0) hit = 0;
			__syncthreads();
			if(rayHit)	   hit = 1;
			__syncthreads();
			if(!hit) break;

			reflectLimit--;

			//The contribution of the reflected ray depends of contribution of previous rays and relfectivity of current material.
			refletSum *= reflectivity;

			//Create reflection ray from intersection point and previously computed reflected direction.
			origin = intersectionPoint;
			dir	   = reflectedDir;
		} else
			break;
	};

	//Store the color to buffer as traditional 8-bit RGB.
	{
		unsigned short x = startX + blockIdx.x*8 + threadIdx.x;
		unsigned short y = startY + blockIdx.y*8 + threadIdx.y;
		unsigned index = __mul24(y,width) + x;

		color.x = min(color.x,1.0f);
		color.y = min(color.y,1.0f);
		color.z = min(color.z,1.0f);

		color *= 255.0f;
		uint3 iColor;
		iColor.x = color.x;
		iColor.y = color.y;
		iColor.z = color.z;

		unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
		buffer[index] = val;
	};
};

float cudaRenderBvh(unsigned int width,unsigned int height,
				    unsigned* cells,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int *buffer,unsigned int *copy,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;

	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};

	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);

	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhPS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhPSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
	};
	
	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	cudaMemcpy(copy,buffer,width*height*4,cudaMemcpyDeviceToHost);

	return elapsedTime;
};

float cudaRenderBvhTexture(unsigned int width,unsigned int height,
				    unsigned* cells,uint4* triangles,cudaArray* vertices,cudaArray *normals,
				    CameraCuda &camera,unsigned lightsCountIn,Light* lightsIn,Material* materials,
				    unsigned int pbo,RenderMode mode) {
	if(width % 8 || height % 8) 
		throw 1;

	float3 eye,upperLeft,dx,dy;
	eye = camera.eye;
	upperLeft = camera.upperLeft;
	dx = camera.dx;
	dy = camera.dy;

	cudaMemcpyToSymbol(lightsCount,&lightsCountIn,sizeof(unsigned)				,0,cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(lights	  , lightsIn	 ,lightsCountIn*sizeof(Light)   ,0,cudaMemcpyHostToDevice);

	cudaError_t error;

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	error = cudaBindTextureToArray(vertexTexture,vertices,channelDesc);
	error = cudaBindTextureToArray(normalTexture,normals ,channelDesc);

	unsigned int *buffer;
	error = cudaGLMapBufferObject((void**)&buffer,pbo);

	dim3 grid;
	dim3 threads(8,8,1);
	unsigned startX,startY;
	unsigned tilesX,tilesY;

	if(width<=512) {
		grid.x = width/8;	grid.y = height/8;
		startX = width;		startY = height;
		tilesX = 1;			tilesY = 1;
	} else if(width<768) {
		grid.x = width/16;	grid.y = height/16;
		startX = width/2;	startY = height/2;
		tilesX = 2;			tilesY = 2;
	} else {
		grid.x = width/32;	grid.y = height/32;
		startX = width/4;	startY = height/4;
		tilesX = 4;			tilesY = 4;
	};

	float elapsedTime;

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start,0);

	switch(mode) {
		case PRIMARY:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhP  <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhPS <<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
		case PRIMARY_SHADOW_REFLECT:
			for(unsigned y=0;y<tilesX;y++)
				for(unsigned x=0;x<tilesY;x++) 
					renderBvhPSR<<<grid,threads>>>(eye,upperLeft,dx,dy,startX*x,startY*y,width,cells,triangles,materials,buffer);
			break;
	};

	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaEventElapsedTime(&elapsedTime,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaGLUnmapBufferObject(pbo);

	cudaUnbindTexture(vertexTexture);
	cudaUnbindTexture(normalTexture);

	return elapsedTime;
};
