/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>

#include "Utils/GLEE/glew.h"
#include <cuda_gl_interop.h>

#include "CameraCuda.h"
#include "Lights.h"
#include "MaterialCuda.h"

#include "RenderMode.h"

#include "vectorMath.cu"


texture<float4, 2, cudaReadModeElementType> vertexTexture;
texture<float4, 2, cudaReadModeElementType> normalTexture;

__constant__ unsigned lightsCount;
__constant__ Light    lights[8];

//Get face normal at given intersection point
__device__ void catchNormal(uint4* triangles, int hitTriangle,float u,float v,float3& normal,unsigned &material) {
	//Normals are stored in 2D texture(Due to the 1D size limitation) with width of 256. Convert linear indexes to 2D coordinates.
	unsigned x1 = triangles[hitTriangle].y &  255;
	unsigned y1 = triangles[hitTriangle].y >> 8;
	unsigned x2 = triangles[hitTriangle].z &  255;
	unsigned y2 = triangles[hitTriangle].z >> 8;
	unsigned x3 = triangles[hitTriangle].w &  255;
	unsigned y3 = triangles[hitTriangle].w >> 8;
	material	= triangles[hitTriangle].x;

	//Read normals from texture
	float3 normal0 = float4to3(tex2D(normalTexture,x1,y1));
	float3 normal1 = float4to3(tex2D(normalTexture,x2,y2));
	float3 normal2 = float4to3(tex2D(normalTexture,x3,y3));

	//The intesection point is given is barycentric coordinates and is therefore weighted average of vertices.
	//The third missing coordinate can be computed from the supplied two as sum must be 1.
	float w = 1.0f - u - v;
	normal0*=w;
	normal1*=u;
	normal2*=v;

	//sum normals and renormalize
	normal0+=normal1;
	normal0+=normal2;
	normalize(normal0);

	normal = normal0;
};

__device__ void shade(float3 &color,bool shadow,unsigned light,
					  Material* materials,unsigned material,
					  float currentVal,float refletSum,
					  float3 &normal,float3 &dir,float3 &reflectedDir) {
	//The standard Phong shading model
	normalize(dir);

	//The ambient light.
	color += materials[material].ambient * lights[light].kAmbient * currentVal;

	//Do diffuse and specular only if not in shadow
	if(!shadow) {
		float coef;

		//Diffuse light
		coef = abs(dot(normal,dir));
		if(coef >= 0.0f)		
			//This is irelevant as coef is now always positive. But in current compiler increase register usage if removed(dont know why);
			color += materials[material].diffuse * lights[light].kDiffuse * coef * currentVal;

		//Specular light
		normalize(reflectedDir);
		coef = -dot(dir,reflectedDir);
		if(coef > 0.0f) {
			coef = powf(coef,materials[material].shinines);
			color += materials[material].specular * lights[light].kDiffuse * coef * refletSum;
		};
	};
};

__device__ void shadePrimary(float3 &color,unsigned light,
							 Material* materials,unsigned material,
						     float3 &normal,float3 &dir,float3 &reflectedDir) {
	//The standard Phong shading model
	normalize(dir);

	//The ambient light.
	color += materials[material].ambient * lights[light].kAmbient;
	
	float coef;
	//Diffuse light
	coef = abs(dot(normal,dir));
	if(coef >= 0.0f)		
		//This is irelevant as coef is now always positive. But in current compiler increase register usage if removed(dont know why);
		color += materials[material].diffuse * lights[light].kDiffuse * coef;

	//Specular light
	normalize(reflectedDir);
	coef = -dot(dir,reflectedDir);
	if(coef > 0.0f) {
		coef   = powf(coef,materials[material].shinines);
		color += materials[material].specular * lights[light].kDiffuse * coef;
	};
};

__device__ void storeInteger(float3 &color,unsigned int* buffer) {
	unsigned short x = blockIdx.x*8 + threadIdx.x;
	unsigned short y = blockIdx.y*8 + threadIdx.y;
	unsigned index = __mul24(y*8,gridDim.x) + x;

	color.x = min(color.x,1.0f);
	color.y = min(color.y,1.0f);
	color.z = min(color.z,1.0f);

	color *= 255.0f;
	uint3 iColor;
	iColor.x = color.x;
	iColor.y = color.y;
	iColor.z = color.z;

	unsigned val = iColor.z<<16 | iColor.y<<8 | iColor.x;
	buffer[index] = val;
	
	/*if(threadIdx.y & 0x01)
		buffer[index] = val;
 
	if(!(threadIdx.y & 0x01))
		buffer[index] = val;*/

};

__device__
void intersectBox(float3 &minimum,float3 &maximum,float3 &eye,float3 &dir,float &tMin,float &tMax) {
	// Kay and Kayjia ray-box intersection test as descibed in "An introduction to Ray Tracing" by A.S.Glassner
	float3 reciprocal = rcp(dir);
	float3 tHitLeft	  = (minimum - eye) * reciprocal;
	float3 tHitRight  = (maximum - eye) * reciprocal;
	
	float3 tHitMin = min(tHitLeft,tHitRight);
	float3 tHitMax = max(tHitLeft,tHitRight);

	tMax = min(min(tHitMax.x,tHitMax.y),tHitMax.z);
	tMin = max(max(tHitMin.x,tHitMin.y),tHitMin.z);
};

#include "grid.cu"
//#include "kdtree.cu"
#include "bvh.cu"


/*
void cudaGenerateOldTest(int width,int height,float3 &minimum, float3 &maximum, float cellSize, uint3 &dim,
				  Cell* cudaGrid,uint3* triangles,float4* vertices,
				  Camera &camera) {
	if(width % 8 || height % 8) 
		throw 1;

	unsigned int* buffer;
	unsigned int* rearange;
	unsigned int* copy;

	cudaMallocHost((void**)&copy    ,width*height*4);
	cudaMalloc	  ((void**)&buffer  ,width*height*4);
	cudaMalloc	  ((void**)&rearange,width*height*4);

	cudaError_t error = cudaGetLastError();
	const char* errors = cudaGetErrorString(error);

	float3 eye,upperLeft,dx,dy;
	camera.prepareCuda(eye,upperLeft,dx,dy);

	cudaEvent_t start,stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start,0);

	dim3 threads(8,8);
	dim3 grid(width / 8, height / 8);

	generate<<<grid,threads>>>(eye,upperLeft,dx,dy,
							   minimum,maximum,cellSize,dim,
							   cudaGrid,triangles,vertices,
							   buffer);

	 //Image coalescing test
	/*dim3 threads(8,8);
	dim3 grid(width / 8, height / 8);

	generateGray<<<grid,threads>>>(buffer);

	threads = dim3(16,8);
	grid	= dim3(width / 16, height /8);

	rearage<<<grid,threads>>>(buffer,rearange);*/
	
	/*dim3 threads(8,8);
	dim3 grid(width / 8, height / 8);

	generateGrayDirect<<<grid,threads>>>(buffer);*/

/*	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	float elapsed;
	cudaEventElapsedTime(&elapsed,start,stop);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaMemcpy(copy,buffer,width*height*4,cudaMemcpyDeviceToHost);

	FILE *outout = fopen("e:\\test.raw","wb");
	fwrite(copy,1,width*height*4,outout);
	fclose(outout);

	cudaFree(buffer);
	cudaFree(rearange);
	cudaFreeHost(copy);
};*/





/*
__global__ void generateGrayOld(unsigned char* buffer) {
	unsigned short x = blockIdx.x*8 + threadIdx.x;
	unsigned short y = blockIdx.y*8 + threadIdx.y;

	unsigned index = (__mul24(y*8,gridDim.x) + x)*3;

	unsigned val = (threadIdx.y*8 + threadIdx.x)*4;
	buffer[index]   = val;
	buffer[index+1] = val;
	buffer[index+2] = val;
};*/
/*
__global__ void generateGrayDirect(unsigned int* buffer) {
	unsigned short x = blockIdx.x*8 + threadIdx.x;
	unsigned short y = blockIdx.y*8 + threadIdx.y;

	unsigned index = (__mul24(y*8,gridDim.x) + x);

	unsigned char val   = (threadIdx.y*8 + threadIdx.x)*4;
	unsigned int  pixel = val<<16 | val<<8 | val;

	if(threadIdx.y & 0x01)
		buffer[index] = pixel;
 
	if(!(threadIdx.y & 0x01))
		buffer[index] = pixel;
};
/*
__global__ void generateGray(unsigned int* buffer) {
	unsigned index = (__mul24(blockIdx.y,gridDim.x) + blockIdx.x)*64;

	unsigned char val   = (threadIdx.y*8 + threadIdx.x)*4;
	unsigned int  pixel = val<<16 | val<<8 | val;

	unsigned offset = (threadIdx.y*8 + threadIdx.x);

	buffer[index + offset] = pixel;
};


__global__ void rearage(unsigned int* buffer,unsigned int* target) {
	__shared__ unsigned int buf[64*2+8];

	unsigned offset = (__mul24(blockIdx.y,gridDim.x) + blockIdx.x)*64*2;
	unsigned index  = threadIdx.y*16 + threadIdx.x;

	unsigned int val;

	val = buffer[offset + index];
	buf[index] = val;

	__syncthreads();

	unsigned pitch = gridDim.x*16;
	offset = __mul24(pitch*8,blockIdx.y) + pitch*threadIdx.y + blockIdx.x*16;

	index = threadIdx.y*8 + threadIdx.x;
	if(threadIdx.x >= 8) index += 64 - 8;

	val = buf[index];
	target[offset + threadIdx.x] = val;
};
*/