#include "StdAfx.h"

#include <wx/image.h>

#include "TextureDisplay.h"

#include <vector_functions.h>

#include "Camera.h"
#include "SceneLights.h"
#include "Material.h"
#include "FormatImport/PlyLoader.h"
#include "Partitoning/UniformGrid.h"

#include "Transformations/SceneNode.h"

#include "MainWindow.h"

#include "Gpu/Wrapper.h"

#include <cuda_gl_interop.h>

const float PI2  = 6.283185307179586476925286766559f;
const float PI2A = 57.295779513082320876798154814105f;
const float STEP = 0.17453292519943295769236907684886f;

extern Resolution RESOLUTIONS[];

BEGIN_EVENT_TABLE(TextureDisplay, wxGLCanvas)
    EVT_SIZE			(TextureDisplay::OnSize)
    EVT_PAINT			(TextureDisplay::OnPaint)
    EVT_ERASE_BACKGROUND(TextureDisplay::OnEraseBackground)
    EVT_MOUSE_EVENTS	(TextureDisplay::OnMouse)
END_EVENT_TABLE()

TextureDisplay::TextureDisplay(	wxWindow *parent, wxWindowID id,
								const wxPoint& pos,
								const wxSize& size, long style,
								const wxString& name)
			   : wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE,name) {
	modelLoaded  = false;
	splashScreen = 0;
	init		 = false;
	renderer	 = 0;
	structure	 = 0;
	mouseDown	 = false;

	isOmni  = true;

	width	= RESOLUTIONS[0].width;
	height	= RESOLUTIONS[0].height;

	camera = new Camera;
	camera->setUpProjection(width,height,1.0f);

	mainAxis = 0;
	angleX = 0.0f;
	//angleX = 1.5707963267948966192313216916398f;
	angleY = 0.0f;
	//angleY = 1.5707963267948966192313216916398f;
	angleZ = 0.0f;

	moveLeft = moveRight = moveFront = moveBack = bankLeft = bankRight = false;

}

TextureDisplay::~TextureDisplay(void) {
	cudaGLUnregisterBufferObject(pbo);
	glDeleteBuffers(1,&pbo);

	delete camera;
	if(renderer) 
		delete renderer;

	if(modelLoaded)
		delete scene;
}

void TextureDisplay::loadTexture(char* filename,Texture& texture) {
	wxImage image(wxString(filename,wxConvISO8859_1));
	unsigned char* data = image.GetData();

	texture.width  = image.GetWidth();
	texture.height = image.GetHeight();

	glGenTextures(1,&texture.texture);
	glBindTexture(GL_TEXTURE_2D,texture.texture);
	glTexEnvf	   ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glTexParameterf( GL_TEXTURE_2D , GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf( GL_TEXTURE_2D , GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D   ( GL_TEXTURE_2D , 0,GL_RGB,texture.width,texture.height,0,GL_RGB,GL_UNSIGNED_BYTE,data );
	glBindTexture(GL_TEXTURE_2D,0);
};

void TextureDisplay::drawCenter(Texture &texture) {
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,texture.texture);

	float left  = -(float)texture.width  / width;
	float up    =  (float)texture.height / height;
	float right = -left;
	float down	= -up;

	glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f); glVertex2f(left ,up);
		glTexCoord2f(1.0f,0.0f); glVertex2f(right,up);
		glTexCoord2f(1.0f,1.0f); glVertex2f(right,down);
		glTexCoord2f(0.0f,1.0f); glVertex2f(left ,down);
	glEnd();

	glBindTexture(GL_TEXTURE_2D,0);
	glEnable(GL_TEXTURE_2D);
};

void TextureDisplay::OnPaint(wxPaintEvent& event) {
	wxPaintDC dc(this);
	SetCurrent();

	if(!init) {
		init = true;
		glewInit();

		glGenBuffers(1, &pbo);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
		glBufferData(GL_PIXEL_UNPACK_BUFFER, RESOLUTIONS[0].width * RESOLUTIONS[0].height * 4, 0, GL_DYNAMIC_DRAW);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

		cudaGLRegisterBufferObject(pbo);

		loadTexture("bitmap/build.png",textureBuild);
		loadTexture("bitmap/load.png" ,textureLoad);
	};


	if(modelLoaded && splashScreen==0) {
		bool moving = moveLeft | moveRight | moveFront | moveBack | bankLeft | bankRight;
		if(moving) {
			float side  = 0.0f;
			float front = 0.0f;

			/*if(moveLeft)  side  -= 0.01f;
			if(moveRight) side  += 0.01f;
			if(moveFront) front -= 0.01f;
			if(moveBack)  front += 0.01f;*/
			if(moveLeft)  side  -= 0.005f;
			if(moveRight) side  += 0.005f;
			if(moveFront) front -= 0.005f;
			if(moveBack)  front += 0.005f;
			if(bankLeft)  angleZ -= 0.1f;
			if(bankRight) angleZ += 0.1f;

			side  *= distance;
			front *= distance;	

			camera->eye -= camera->moveSide  * side;
			camera->eye -= camera->moveFront * front;

//			printf("Camera x=%f y=%f z=%f\n",camera->eye.x,camera->eye.y,camera->eye.z);

			if(bankLeft | bankRight) {
				recomputeCamera();
			};
		};
	
		SceneLights lights;
		lights.push_back(Light());
		lights.back().possition = make_float3(0.0f,0.0f,1.0f);
		lights.back().kAmbient  = make_float3(0.2f,0.2f,0.2f);
		lights.back().kDiffuse  = make_float3(0.8f,0.8f,0.8f);
		lights.back().possition = make_float3(camera->eye.x,camera->eye.y,camera->eye.z);

		float time = renderer->render(*camera,lights,PRIMARY);
		((MainWindow*)this->GetParent())->setRenderTime(time);
	
		glBindBuffer (GL_PIXEL_UNPACK_BUFFER,pbo);
		glPixelZoom( 1.0, -1.0 );
		glRasterPos2f(-1.0f,1.0f);
		glDrawPixels(width,height,GL_RGBA,GL_UNSIGNED_BYTE,0);
		glBindBuffer (GL_PIXEL_UNPACK_BUFFER,0);

		glFlush();
		SwapBuffers();

		if(moving)
			Refresh(false);
	} else {
		glClear(GL_COLOR_BUFFER_BIT);

		if(splashScreen == 1) drawCenter(textureLoad);
		if(splashScreen == 2) drawCenter(textureBuild);

		glFlush();
		SwapBuffers();
	};
};

void TextureDisplay::recomputeCamera() {
	if(isOmni) {
		SPVector axis  (cosf(angleX),0.0f,-sinf(angleX),0.0f);
		SPVector eye   (distance * sinf(angleX),0.0f,distance * cosf(angleX),1.0f);
		SPVector up	   (0.0f,1.0f,0.0f,0.0f);
		SPVector target(0.0f,0.0f,0.0f,1.0f);

		Transformation rotate;
		rotate.IdentityMatrix();
		rotate.rotate(axis.x,axis.y,axis.z,angleY);

		eye *= rotate;
		up  *= rotate;

		switch(mainAxis) {
			case 0: eye.vec    = _mm_shuffle_ps(eye.vec	  ,eye.vec		,_MM_SHUFFLE(3,2,0,1));
					target.vec = _mm_shuffle_ps(target.vec,target.vec	,_MM_SHUFFLE(3,2,0,1));
					up.vec	   = _mm_shuffle_ps(up.vec	  ,up.vec		,_MM_SHUFFLE(3,2,0,1)); break;
			case 1: eye.vec    = _mm_shuffle_ps(eye.vec	  ,eye.vec		,_MM_SHUFFLE(3,1,2,0));
					target.vec = _mm_shuffle_ps(target.vec,target.vec	,_MM_SHUFFLE(3,1,2,0));
					up.vec	   = _mm_shuffle_ps(up.vec	  ,up.vec		,_MM_SHUFFLE(3,1,2,0)); break;
		};

		camera->setLook(eye,target,up);
	} else {
		SPVector front (0.0f		, 0.0f		  ,-1.0f,0.0f);
		SPVector up	   (sinf(angleZ), cosf(angleZ), 0.0f,0.0f);
		SPVector side  (cosf(angleZ),-sinf(angleZ), 0.0f,0.0f);

		switch(mainAxis) {
			case 0: side.vec  = _mm_shuffle_ps(side.vec	,side.vec	,_MM_SHUFFLE(3,0,2,1));
					front.vec = _mm_shuffle_ps(front.vec,front.vec	,_MM_SHUFFLE(3,0,2,1));
					up.vec	  = _mm_shuffle_ps(up.vec	,up.vec		,_MM_SHUFFLE(3,0,2,1)); break;
			case 1: side.vec  = _mm_shuffle_ps(side.vec	,side.vec	,_MM_SHUFFLE(3,1,2,0));
					front.vec = _mm_shuffle_ps(front.vec,front.vec	,_MM_SHUFFLE(3,1,2,0));
					up.vec	  = _mm_shuffle_ps(up.vec	,up.vec		,_MM_SHUFFLE(3,1,2,0)); break;
		};

		
		Transformation rotateUp;
		rotateUp.IdentityMatrix();
		rotateUp.rotate(up.x,up.y,up.z,angleX);

		front *= rotateUp;
		side  *= rotateUp;

		Transformation rotateSide;
		rotateSide.IdentityMatrix();
		rotateSide.rotate(side.x,side.y,side.z,angleY);

		front *= rotateSide;
		up	  *= rotateSide;

		SPVector eye   (camera->eye);
		front+=eye;

//		printf("Angle x=%f y=%f\n",angleX,angleY);

		camera->setLook(eye,front,up);
	};
};

void TextureDisplay::OnMouse(wxMouseEvent& event) {
	if(event.LeftDown()) {
		mouseDown = true;
		lastX	  = event.GetX();
		lastY	  = event.GetY();

		clickAngleX = angleX;
		clickAngleY = angleY;
	} 
	else if(event.LeftUp()) {
		mouseDown = false;
	};

	if(mouseDown) {
		wxCoord curX = event.GetX();
		wxCoord curY = event.GetY();

	 	angleX = clickAngleX + (float)(curX - lastX)*0.01f;
		angleY = clickAngleY + (float)(curY - lastY)*0.01f;

		if(angleX <  0.0f) angleX+=PI2;
		if(angleX >= PI2 ) angleX-=PI2;
		if(angleY <  0.0f) angleY+=PI2;
		if(angleY >= PI2 ) angleY-=PI2;

		recomputeCamera();
		Refresh(false);
	};
};

void TextureDisplay::setMainAxis(unsigned axis) {
	mainAxis = axis;
	recomputeCamera();
	Refresh(false);
};

void TextureDisplay::loadModel(const char* filename) {
	if(modelLoaded) {
		modelLoaded = false;
		delete renderer;
		delete scene;
	};

	splashScreen = 1;
	Refresh(false);
	GetParent()->Update();

	scene = SceneLoader::instance()->loadScene(filename,true);

	splashScreen = 2;
	Refresh(false);
	GetParent()->Update();
	
	switch(structure) {
		case 0:  renderer = new CudaRenderGrid(width,height,new UniformGrid(*scene)	,pbo); break;
		case 1:  renderer = new CudaRenderKd  (width,height,new KdTree	   (*scene)	,pbo); break;
		case 2:  renderer = new CudaRenderBvh (width,height,new Bvh		   (*scene)	,pbo); break;
	};

	distance = renderer->getRenderDistance()*1.5f;
	recomputeCamera();

	modelLoaded  = true;
	splashScreen = 0;
	Refresh(false);
};

void TextureDisplay::setNewStructure(unsigned newStructure) {
	structure = newStructure;

	if(modelLoaded) {
		splashScreen = 2;
		Refresh(false);
		GetParent()->Update();

		delete renderer;

		switch(structure) {
			case 0:  renderer = new CudaRenderGrid(width,height,new UniformGrid(*scene)	,pbo); break;
			case 1:  renderer = new CudaRenderKd  (width,height,new KdTree	   (*scene)	,pbo); break;
			case 2:  renderer = new CudaRenderBvh (width,height,new Bvh		   (*scene)	,pbo); break;
		};

		splashScreen = 0;
		Refresh(false);
	};
};

void TextureDisplay::setNewResolution(unsigned n) {
	width	= RESOLUTIONS[n].width;
	height	= RESOLUTIONS[n].height;

	cudaGLUnregisterBufferObject(pbo);

	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
	glBufferData(GL_PIXEL_UNPACK_BUFFER, width * height * 4, 0, GL_DYNAMIC_DRAW);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

	cudaGLRegisterBufferObject(pbo);

	camera->setUpProjection(width,height,1.0f);

	if(renderer) {
		renderer->resize(width,height);
		recomputeCamera();
	};

	glViewport(0,0,width,height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0f,1.0f,-1.0f,1.0f,0.0f,10.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
};

void TextureDisplay::keyDown(Direction key) {
	if(isOmni) return;

	switch(key) {
		case MOVE_LEFT  : moveLeft  = true; break;
		case MOVE_RIGHT : moveRight = true; break;
		case MOVE_FRONT : moveFront = true; break;
		case MOVE_BACK	: moveBack  = true; break;
		case BANK_LEFT	: bankLeft  = true; break;
		case BANK_RIGHT	: bankRight = true; break;
	};
	Refresh(false);
};

void TextureDisplay::keyUp(Direction key) {
	switch(key) {
		case MOVE_LEFT  : moveLeft  = false; break;
		case MOVE_RIGHT : moveRight = false; break;
		case MOVE_FRONT : moveFront = false; break;
		case MOVE_BACK	: moveBack  = false; break;
		case BANK_LEFT	: bankLeft  = false; break;
		case BANK_RIGHT	: bankRight = false; break;
	};
};

void TextureDisplay::setViewMode(bool omni) {
	isOmni = omni;
};

void TextureDisplay::OnSize(wxSizeEvent& event) {
	wxGLCanvas::OnSize(event);
};
  
void TextureDisplay::OnEraseBackground(wxEraseEvent& event) {

};