/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <algorithm>
#include <float.h>

#include <limits>

#include <channel_descriptor.h>
#include <driver_functions.h>
#include <cuda_runtime_api.h>

#include <assert.h>

#include "utils/iterator.h"

#include "Bvh.h"

#include <time.h>

using sse::operator+;
using sse::operator+=;
using sse::operator-;
using sse::operator*;
using sse::operator/;
using sse::operator<;
using sse::operator&;
using sse::operator==;
using sse::operator|;

using std::min;
using std::max;

const unsigned X=0;
const unsigned Y=1;
const unsigned Z=2;

Bvh::Bvh(Scene &scene) {
	vertexSum = 0;
	facesSum  = 0;

	/////////////////////////////
	// Find grid extents. The individual object's bounding boxes are transformed
	// into world space and the minimum and maximum is found;
	/////////////////////////////

	//Initialize minimum and maximum to oposite limits to ensure first comparation results in assigning the value
	__m128 minimum,maximum;
	minimum = minimumInit;
	maximum = maximumInit;

	size_t sceneSize = scene.size();
	for(size_t i=0;i<sceneSize;i++) {
		//Prepare total number of faces and vertieces in grid
		vertexSum += scene[i].object->vertexCount;
		facesSum  += scene[i].object->facesCount;

		Vertex &min = scene[i].object->minimum;
		Vertex &max = scene[i].object->maximum;

		//Transform all corners of bounding box into world space and then check for minimality/maximality
		for(int j=0;j<8;j++) {
			SPVector v1;
			v1.x = j & 0x01 ? min.x : max.x;
			v1.y = j & 0x02 ? min.y : max.y;
			v1.z = j & 0x04 ? min.z : max.z;
			v1.w = 1.0f;

			v1 *= *scene[i].transformation;

			minimum = _mm_min_ps(minimum,v1.vec);
			maximum = _mm_max_ps(maximum,v1.vec);
		};
	};

	ALIGN16 float4 store;
	sse::store(minimum,&store);
	mini.x = store.x;
	mini.y = store.y;
	mini.z = store.z;

	sse::store(maximum,&store);
	maxi.x = store.x;
	maxi.y = store.y;
	maxi.z = store.z;

	int curVertex = 0;
	int curFace   = 0;
	unsigned faceOffset   = 0;
	unsigned vertexOffset = 0;

	transformed		= new Vertex[vertexSum];
	normals			= new Vertex[vertexSum];

	//Triangle *triangles = new Triangle[facesSum];
	triangles		= new Triangle[facesSum];

	boxes			= new BoundingBox[facesSum];
	centroids	 	= new Vertex[facesSum];

	clock_t start,stop;
	start = clock();

	static const __m128 ONE_THIRD  = {1.0f/3.0f,1.0f/3.0f,1.0f/3.0f,1.0f/3.0f};
	static const __m128i maskFirst = sse::set(0,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF);

	for(size_t i=0;i<sceneSize;i++) {
		unsigned facesCount  = scene[i].object->facesCount;
		unsigned vertexCount = scene[i].object->vertexCount;

		__m128i vertexSumSse = _mm_setr_epi32(0,vertexOffset,vertexOffset,vertexOffset);

		Vertex*			__restrict vertices	 = scene[i].object->vertices;
		Vertex*			__restrict normalsIn = scene[i].object->normal;
		Triangle*		__restrict faces	 = scene[i].object->faces;
		Transformation* __restrict matrix	 = scene[i].transformation;

		for(unsigned j=0;j<vertexCount;j++) {
			transformed[curVertex].vec = ( vertices[j] * *matrix).vec;
			normals[curVertex++]	   = normalsIn[j];
		};

		std::vector<unsigned> &materialIndexes = scene[i].replacementMaterials.size() ? scene[i].replacementMaterials : scene[i].object->materials;

		for(unsigned j=0;j<facesCount;j++) {
			__m128i triangle = sse::load<__m128i>(&faces[j]);						//Load the triangle
			__m128i material = sse::pack(materialIndexes[sse::extract(triangle)]);	//Change the object material number to global material number.

			//Mask the material into the triangle
			triangle = triangle & maskFirst;
			triangle = triangle | material;

			//Convert object triangle indices to global indices
			triangle += vertexSumSse;

			sse::store(triangle,&triangles[j + faceOffset]);
		};

		faceOffset   += facesCount;
		vertexOffset += vertexCount;
	};
	
	for(unsigned i=0;i<facesSum;i++) {
		__m128 a,b,c;
		a = sse::load<__m128>(&transformed[triangles[i].a]);
		b = sse::load<__m128>(&transformed[triangles[i].b]);
		c = sse::load<__m128>(&transformed[triangles[i].c]);

		boxes[i].minimum = _mm_min_ps(_mm_min_ps(b,c),a);
		boxes[i].maximum = _mm_max_ps(_mm_max_ps(b,c),a);

		a = a+b;
		a = a+c;
		a = a*ONE_THIRD;

		//centroids[i].vec = a;
		_mm_stream_ps(&centroids[i].x,a);
	};

	_MM_SET_ROUNDING_MODE(_MM_ROUND_TOWARD_ZERO);

	cellPrimitives = new TailPool<unsigned>(facesSum*3);
	cellTriangles  = new GrowArray<Triangle>(facesSum);
	cells		   = new GrowArray<BvhCell>(facesSum/4);
	cells->getNext();
	cells->getNext();

	buildBvh<true>(minimum,maximum,(float*)&minimum,(float*)&maximum,triangles,0,0,facesSum,0);

	delete[] boxes;
	delete[] centroids;
	delete cellPrimitives;

	stop  = clock();
//	printf("%f",(double)(stop-start)/CLOCKS_PER_SEC);
}

Bvh::~Bvh(void) {
	delete cellTriangles;
	delete cells;

	if(videoAllocated) {
		cudaFree(savedTriangles);
		cudaFree(savedTree);
		cudaFreeArray(savedVertices);
		cudaFreeArray(savedNormals);
	};
}

const unsigned RATIO   = 6;
const unsigned BIN_MAX = 128;
const unsigned BIN_MIN = 8;

struct Bin:public AllignedS {
	__m128 minimum;
	__m128 maximum;
	unsigned count;
	float	 minCentroid;
	uint2	 allignFill;

	void merge(Bin& second);
};

void Bin::merge(Bin& second) {
	if(second.count) {
		count += second.count;
		minCentroid = min(minCentroid,second.minCentroid);
		minimum		= sse::min(minimum,second.minimum);
		maximum		= sse::max(maximum,second.maximum);
	};
};

void clearBin(Bin* bin,unsigned count) {
	for(unsigned i=0;i<count;i++) {
		bin[i].minimum		= minimumInit;
		bin[i].maximum		= maximumInit;
		bin[i].count		= 0;
		bin[i].minCentroid  = FLT_MAX;
	};
};

template<bool initial>
void Bvh::buildBvh(__m128 minimum,__m128 maximum,float* centroidMinf,float* centroidMaxf,
				   Triangle* faces,unsigned node,unsigned* indices,unsigned count,unsigned depth) {
	//static const __m128 MARGIN    = {0.001f,0.001f,0.001f,0.001f};
	//static const uint4  ABS		  = {0x7FFFFFFF,0x7FFFFFFF,0x7FFFFFFF,0x7FFFFFFF};

	ALIGN16 static Bin bins[129][3];
	static float leftPriceX[129];
	static float leftPriceY[129];
	static float leftPriceZ[129];

	if(node == 32160 ) 
		node=node;

	if(count <= 4) {
		ALIGN16 float fMinimum[4];
		ALIGN16 float fMaximum[4];
		sse::store(minimum,fMinimum);
		sse::store(maximum,fMaximum);

		BvhCell *cell	 = (*cells)[node];
		cell->count		 = count;
		cell->minimum.x  = fMinimum[X];
		cell->minimum.y  = fMinimum[Y];
		cell->minimum.z  = fMinimum[Z];
		cell->maximum.x  = fMaximum[X];
		cell->maximum.y  = fMaximum[Y];
		cell->maximum.z  = fMaximum[Z];
		
		unsigned triangleStart = cellTriangles->getMore(4);
		Triangle* start = (*cellTriangles)[triangleStart];
		cell->children  = triangleStart;

		for(unsigned i=0;i<count;i++)
			start[i] = faces[indices[i]];
		for(unsigned i=count;i<4;i++) {
			start[i].a = -1;
			start[i].b = -1;
			start[i].c = -1;
		};
		
		return;
	};

	unsigned binCount = count/RATIO;
	binCount = max(binCount,BIN_MIN);
	binCount = min(binCount,BIN_MAX);

	clearBin(bins[0],binCount*3+3);

	float cellSurface;
	
	{
		__m128 size = maximum - minimum;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&cellSurface);
		cellSurface = 1.0f/cellSurface;
	};
	
	__m128 centroidMin = sse::load<__m128>(centroidMinf);
	__m128 centroidMax = sse::load<__m128>(centroidMaxf);

	//Prepare nodes extent size for bin computation. Must be checked for zero sized extent(infinity after division).
	//Change zero extent to infinity so it is zero after deivision.
	ALIGN16 float4 prepareSize;
	prepareSize.x = centroidMaxf[0] - centroidMinf[0]; if(prepareSize.x == 0.0f) prepareSize.x = std::numeric_limits<float>::infinity();
	prepareSize.y = centroidMaxf[1] - centroidMinf[1]; if(prepareSize.y == 0.0f) prepareSize.y = std::numeric_limits<float>::infinity();
	prepareSize.z = centroidMaxf[2] - centroidMinf[2]; if(prepareSize.z == 0.0f) prepareSize.z = std::numeric_limits<float>::infinity();
	prepareSize.w = std::numeric_limits<float>::infinity();

	__m128 size		 = sse::load<__m128>(&prepareSize);
	__m128 binCount4 = sse::set4(binCount);
	size = binCount4 / size;
	
	for(unsigned i=0;i<count;i++) {
		ALIGN16 uint4  iCenter;
		ALIGN16 float4 fCenter;

		__m128 minExtent,maxExtent,centroid;

		//Load precomputed bounding box and centroid.
		if(initial) {
			minExtent = boxes[i].minimum;
			maxExtent = boxes[i].maximum;
			centroid  = centroids[i].vec;
		} else {
			minExtent = boxes[indices[i]].minimum;
			maxExtent = boxes[indices[i]].maximum;
			centroid  = centroids[indices[i]].vec;

			//Speed up algoritm by prefetching next step data
			_mm_prefetch((char*)&boxes[indices[i+1]],_MM_HINT_T2);
			_mm_prefetch((char*)&centroids[indices[i+1]],_MM_HINT_T1);
		};

		sse::store(centroid,&fCenter);

		//Compute the bin coordinates from centroid possition
		centroid = centroid - centroidMin;
		centroid = centroid * size;

		sse::store(sse::f4i(centroid),&iCenter);

		//if(iCenter.x < 0 || iCenter.y < 0 || iCenter.z < 0) throw 1;
		//if(iCenter.x > binCount || iCenter.y > binCount || iCenter.z > binCount) throw 1;

		//Update bins for every axis.
		Bin *current;
		current=&bins[iCenter.x][X];
		current->count++;
		current->minCentroid = min(current->minCentroid,fCenter.x);
		current->minimum	 = sse::min(current->minimum,minExtent);
		current->maximum	 = sse::max(current->maximum,maxExtent);

		current=&bins[iCenter.y][Y];
		current->count++;
		current->minCentroid = min(current->minCentroid,fCenter.y);
		current->minimum	 = sse::min(current->minimum,minExtent);
		current->maximum	 = sse::max(current->maximum,maxExtent);

		current=&bins[iCenter.z][Z];
		current->count++;
		current->minCentroid = min(current->minCentroid,fCenter.z);
		current->minimum	 = sse::min(current->minimum,minExtent);
		current->maximum	 = sse::max(current->maximum,maxExtent);
	};

	//Some centroids might be placed exactly at the right bin boundary, therefore overflowing bin array.
	//Instad of checking in the loop, increase the size of array by one, allow the owerflow and merge the exces
	//bin to the proper last one here.
	bins[binCount-1][X].merge(bins[binCount][X]);
	bins[binCount-1][Y].merge(bins[binCount][Y]);
	bins[binCount-1][Z].merge(bins[binCount][Z]);

	unsigned countX=0,countY=0,countZ=0;
	__m128	 minX=minimumInit,minY=minimumInit,minZ=minimumInit;
	__m128	 maxX=maximumInit,maxY=maximumInit,maxZ=maximumInit;
	
	//First left to right pass over the bins.
	for(unsigned i=0;i<binCount;i++) {
		float surfaceX,surfaceY,surfaceZ;

		//Update lest side object count;
		countX += bins[i][X].count;
		countY += bins[i][Y].count;
		countZ += bins[i][Z].count;

		//Update the bounding boxes
		minX = sse::min(minX,bins[i][X].minimum);
		maxX = sse::max(maxX,bins[i][X].maximum);
		minY = sse::min(minY,bins[i][Y].minimum);
		maxY = sse::max(maxY,bins[i][Y].maximum);
		minZ = sse::min(minZ,bins[i][Z].minimum);
		maxZ = sse::max(maxZ,bins[i][Z].maximum);

		//Computer bounding box surface area for all three axis.
		__m128 size;
		size = maxX - minX;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&surfaceX);

		size = maxY - minY;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&surfaceY);

		size = maxZ - minZ;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&surfaceZ);

		//Save the data in array for use in second pass.
		leftPriceX[i] = countX*surfaceX;
		leftPriceY[i] = countY*surfaceY;
		leftPriceZ[i] = countZ*surfaceZ;
	};

	countX=0,countY=0,countZ=0;
	minX=minY=minZ=minimumInit;
	maxX=maxY=maxZ=maximumInit;
	
	float bestX = FLT_MAX;
	float bestY = FLT_MAX;
	float bestZ = FLT_MAX;
	unsigned bestCount[3];
	unsigned bestIndex[3];

	//The second right to left pass over the bins.
	for(int i=binCount-1;i>=0;i--) {
		float surfaceX,surfaceY,surfaceZ;

		countX += bins[i][X].count;
		countY += bins[i][Y].count;
		countZ += bins[i][Z].count;

		minX = sse::min(minX,bins[i][X].minimum);
		maxX = sse::max(maxX,bins[i][X].maximum);
		minY = sse::min(minY,bins[i][Y].minimum);
		maxY = sse::max(maxY,bins[i][Y].maximum);
		minZ = sse::min(minZ,bins[i][Z].minimum);
		maxZ = sse::max(maxZ,bins[i][Z].maximum);

		__m128 size;
		size = maxX - minX;												
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));	//x*y,y*z,z*x
		sse::sum(size);													//xy+yz+xz
		sse::store1(size + size,&surfaceX);								//2*(xy+yz+xz);

		size = maxY - minY;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&surfaceY);

		size = maxZ - minZ;
		size = size * _mm_shuffle_ps(size,size,_MM_SHUFFLE(3,0,2,1));
		sse::sum(size);
		sse::store1(size + size,&surfaceZ);

		//Using the data from first pass, compute the spliting cost.
		float priceX = 1.0f + cellSurface*(countX*surfaceX + leftPriceX[i]);
		float priceY = 1.0f + cellSurface*(countY*surfaceY + leftPriceY[i]);
		float priceZ = 1.0f + cellSurface*(countZ*surfaceZ + leftPriceZ[i]);
		//Update best possition info.
		if(priceX < bestX) { bestX = priceX; bestIndex[0]=i;bestCount[0]=countX;};
		if(priceY < bestY) { bestY = priceY; bestIndex[1]=i;bestCount[1]=countY;};
		if(priceZ < bestZ) { bestZ = priceZ; bestIndex[2]=i;bestCount[2]=countZ;};
	};

	ALIGN16 float fMinimum[4];
	ALIGN16 float fMaximum[4];
	sse::store(minimum,fMinimum);
	sse::store(maximum,fMaximum);

	//Prevent trying division on already zero size extent.
	if(centroidMaxf[0] - centroidMinf[0] == 0) bestX=FLT_MAX;
	if(centroidMaxf[1] - centroidMinf[1] == 0) bestY=FLT_MAX;
	if(centroidMaxf[2] - centroidMinf[2] == 0) bestZ=FLT_MAX;

	unsigned minIndex = bestX < bestY ? 
					    bestX < bestZ ? 0 : 2 :
						bestY < bestZ ? 1 : 2;
	unsigned bestBin  = bestIndex[minIndex];
	float	 boundary = bins[bestBin][minIndex].minCentroid; //Use the minimal centroid in bin for correct division into the children.

	//Bins with no triangles have no minimal centroid and can be selected as best spliting possition. 
	//In that case, place the spliting plane possition in the middle of the bin.
	if(boundary == FLT_MAX)
		boundary = centroidMinf[minIndex] + ((float)bestBin+0.5f)*(centroidMaxf[minIndex] - centroidMinf[minIndex])/binCount;

	unsigned rightCount = bestCount[minIndex];
	unsigned leftCount  = count - rightCount;

	unsigned leftIndex  = 0;
	unsigned rightIndex = 0;

	unsigned* rightFaces = cellPrimitives->getMem(rightCount);
	unsigned* leftFaces  = cellPrimitives->getMem(leftCount);

	__m128 leftBoundMin	 = minimumInit;
	__m128 leftBoundMax  = maximumInit;
	__m128 rightBoundMin = minimumInit;
	__m128 rightBoundMax = maximumInit;

	//Reconstruct the children nodes bounding boxes.
	for(unsigned i=0;i<=bestBin;i++) {
		leftBoundMin = sse::min(leftBoundMin,bins[i][minIndex].minimum);
		leftBoundMax = sse::max(leftBoundMax,bins[i][minIndex].maximum);
	};
	for(unsigned i=binCount-1;i>=bestBin;i--) {
		rightBoundMin = sse::min(rightBoundMin,bins[i][minIndex].minimum);
		rightBoundMax = sse::max(rightBoundMax,bins[i][minIndex].maximum);
	};

	__m128 leftCentroidMin  = minimumInit;
	__m128 leftCentroidMax  = maximumInit;
	__m128 rightCentroidMin = minimumInit;
	__m128 rightCentroidMax = maximumInit;

	//Place the triangles into the children and update the centorid extent for better binning precision.
	for(unsigned i=0;i<count;i++) {
		unsigned face = initial ? i : indices[i];

		float  centroidAxis  = ((float*)&centroids[face])[minIndex];
		__m128 centroidAll   =			 centroids[face].vec;

		if(centroidAxis < boundary) {
			leftFaces[leftIndex++]   = face;
			leftCentroidMin = sse::min(leftCentroidMin,centroidAll);
			leftCentroidMax = sse::max(leftCentroidMax,centroidAll);
		} else {
			rightFaces[rightIndex++] = face;
			rightCentroidMin = sse::min(rightCentroidMin,centroidAll);
			rightCentroidMax = sse::max(rightCentroidMax,centroidAll);
		};
	};

	//Allocate cells structures
	unsigned newCell    = node;
	unsigned leftChild  = cells->getNext();
	unsigned rightChild = cells->getNext();

	BvhCell *cell = (*cells)[newCell];
	cell->children   = leftChild;
	cell->count		 = 0;
	cell->minimum.x  = fMinimum[X];
	cell->minimum.y  = fMinimum[Y];
	cell->minimum.z  = fMinimum[Z];
	cell->maximum.x  = fMaximum[X];
	cell->maximum.y  = fMaximum[Y];
	cell->maximum.z  = fMaximum[Z];

	buildBvh<false>(leftBoundMin,leftBoundMax,(float*)&leftCentroidMin,(float*)&leftCentroidMax,
					triangles,leftChild,leftFaces,leftCount,depth+1);

	cellPrimitives->returnMem(leftCount);

	buildBvh<false>(rightBoundMin,rightBoundMax,(float*)&rightCentroidMin,(float*)&rightCentroidMax,
					triangles,rightChild,rightFaces,rightCount,depth+1);

	cellPrimitives->returnMem(rightCount);
};

void Bvh::prepareCuda(float3 &minimum, float3 &maximum,unsigned* &cellsOut,uint4* &triangles,cudaArray* &vertices,cudaArray* &normalsOut) {
	minimum = mini;
	maximum = maxi;

	cudaChannelFormatDesc channelDesc;
	cudaExtent			  gridSize;

	cudaError_t error;

	cellsOut	= 0;
	vertices	= 0;
	normalsOut	= 0;
	triangles	= 0;

	try {

	unsigned sideX = 256;
	unsigned sideY = vertexSum/sideX + 1;

	channelDesc = cudaCreateChannelDesc<float4>();
	gridSize	= make_cudaExtent(sideX, sideY, 0);

	SAFE_ALLOC(cudaMallocArray(&vertices, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(vertices,0,0,transformed,vertexSum*sizeof(float4),cudaMemcpyHostToDevice);

	SAFE_ALLOC(cudaMallocArray(&normalsOut, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(normalsOut,0,0,normals,vertexSum*sizeof(float4),cudaMemcpyHostToDevice);

/////////////////////////////

	size_t pitch;

	SAFE_ALLOC(cudaMallocPitch((void**)&cellsOut, &pitch, 64, (cells->size()+1)*sizeof(BvhCell)/64));
	cudaMemcpy(cellsOut,(*cells)[0],cells->size()*sizeof(BvhCell),cudaMemcpyHostToDevice);
	
	SAFE_ALLOC(cudaMallocPitch((void**)&triangles, &pitch, 64, (cellTriangles->size()+1)*sizeof(Triangle)/64));
	cudaMemcpy(triangles,(*cellTriangles)[0],cellTriangles->size()*sizeof(Triangle),cudaMemcpyHostToDevice);

	} catch(OutOfMemory) {
		if(triangles)  cudaFree(triangles);
		if(cellsOut)   cudaFree(cellsOut);
		if(vertices)   cudaFreeArray(vertices);
		if(normalsOut) cudaFreeArray(normalsOut);
		throw OutOfMemory();
	};

	videoAllocated  = true;
	savedTree		= cellsOut;
	savedTriangles  = triangles;
	savedVertices	= vertices;
	savedNormals	= normalsOut;
};