/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#include "Utils/Memory.h"

#include "Object.h"
#include "Transformations/SceneNode.h"

struct BvhCell {
	float3	 minimum;
	float3   maximum;
	unsigned children;
	unsigned count;
};

struct BoundingBox:public AllignedS {
	__m128 minimum;
	__m128 maximum;
};

class Bvh {
public:
	float3 mini;	//!< Minimum extent of the root node, prepared for CUDA.
	float3 maxi;	//!< Maximum extent of the root node, prepared for CUDA.
	int vertexSum;	//!< Total number of vertices of objects included in the grid.
	int facesSum;	//!< Total number of faces of objects included in the grid.

	Vertex		*transformed;	//!< Vertices of objects in grid, transformed into world space.
	Vertex		*normals;		//!< Normals associated with vertices. Has one normal for each vertex.
	Triangle    *triangles;		//!< Temporal triangle indexes collected from all scene objects.
								/*!< \warning This is here for debuging and could be removed in later versions.*/

	BoundingBox *boxes;			//!< Temporal storage of triangle bounding boxes.
	Vertex		*centroids;		//!< Temporal storage of triangle centroids.

	GrowArray<BvhCell>  *cells;				//!< Nodes of the tree	
	GrowArray<Triangle> *cellTriangles;		//!< Faces of the tree. 
											/*!< Referenced from tree nodes. Has size of total number of triangles*/
	TailPool<unsigned>  *cellPrimitives;	//!< Temporal indicies used to pass relevant faces to children nodes.

	bool		videoAllocated;	//!< True if video memory is allocated.
	unsigned*	savedTree;		//!< Saved pointer for video memory dealocation.
	uint4*	    savedTriangles; //!< Saved pointer for video memory dealocation.
	cudaArray*  savedVertices;  //!< Saved pointer for video memory dealocation.
	cudaArray*  savedNormals;   //!< Saved pointer for video memory dealocation.

	//! Build node using aproximate min-max binning
	/**
	*	@tparam initial		True if building the root node. In that case, the indices are ignored and the faces are accessed directly.
	*	@param minimum		The minimum extent of the new node.
	*	@param maximum		The maximum extent of the new node.
	*	@param centroidMin	The minimum extent of the node triangles centroids.
	*	@param centroidMax  The maximum extent of the node triangles centroids.
	*	@param faces		The source triangles colectected from all scene objects.
	*	@param node			The node number assigned to the new node.
	*   @param indices		The indices to the triangles witch belongs to this node
	*	@param count		The number of triangles for this node
	*	@param depth		The depth of the cell in the tree
	*/
	template<bool initial>
	void buildBvh(__m128 minimum,__m128 maximum,float* centroidMin,float* centroidMax,
				  Triangle* faces,unsigned node,unsigned* indices,unsigned count,unsigned depth);

public:
	Bvh(Scene &scene);
	~Bvh(void);

	//! Prepare tree for use in CUDA
	/*!
	*	Copy the tree data to video card's linear memory. Uses texture to store the 
	*	vertices,normals and linear memory to store nodes and triangles. 
	*
	*   @param minimum		Returns minimum extent of the root node.
	*	@param maximum		Returns maximum extent of the root node.
	*	@param cells		Pointer to tree nodes in video memory
	*	@param triangles	Pointer to triangles in video memory
	*	@param vertices		Pointer to vertices in video memory
	*	@param normalsOut   Pointer to normals in video memory
	*/
	void prepareCuda(float3 &minimum, float3 &maximum,unsigned* &cells,uint4* &triangles,
					 cudaArray* &vertices,cudaArray* &normalsOut);
};
