/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <algorithm>
#include <float.h>

#include <channel_descriptor.h>
#include <driver_functions.h>
#include <cuda_runtime_api.h>
#include <cuda.h>

#include <assert.h>

#include "utils/iterator.h"

#include "KdTree.h"

#include <time.h>

extern void dumpMainHeap();

using sse::operator+;
using sse::operator+=;
using sse::operator-;
using sse::operator*;
using sse::operator/;
using sse::operator<;
using sse::operator&;
using sse::operator==;
using sse::operator|;

using std::min;
using std::max;

__m128 SMALL = { 0.000001f, 0.000001f, 0.000001f, 0.000001f};

KdTree::KdTree(Scene &scene) {
	videoAllocated = false;
	vertexSum = 0;
	facesSum  = 0;

	/////////////////////////////
	// Find grid extents. The individual object's bounding boxes are transformed
	// into world space and the minimum and maximum is found;
	/////////////////////////////

	//Initialize minimum and maximum to oposite limits to ensure first comparation results in assigning the value
	__m128 minimum,maximum;
	minimum = minimumInit;
	maximum = maximumInit;

	size_t sceneSize = scene.size();
	for(size_t i=0;i<sceneSize;i++) {
		//Prepare total number of faces and vertieces in grid
		vertexSum += scene[i].object->vertexCount;
		facesSum  += scene[i].object->facesCount;

		Vertex &min = scene[i].object->minimum;
		Vertex &max = scene[i].object->maximum;

		//Transform all corners of bounding box into world space and then check for minimality/maximality
		for(int j=0;j<8;j++) {
			SPVector v1;
			v1.x = j & 0x01 ? min.x : max.x;
			v1.y = j & 0x02 ? min.y : max.y;
			v1.z = j & 0x04 ? min.z : max.z;
			v1.w = 1.0f;

			v1 *= *scene[i].transformation;

			minimum = sse::min(minimum,v1.vec);
			maximum = sse::max(maximum,v1.vec);
		};
	};

	__m128 margin = maximum - minimum;
	margin  = margin * SMALL;
	minimum = minimum - margin;
	maximum = maximum + margin;

	//Copy the minimum and maximum extent for later use in prepareCuda. We need only 3 values from the 4 element
	//sse register. Therefore this rather cumbersome element by element copy.
	ALIGN16 float4 store;
	sse::store(minimum,&store);
	mini.x = store.x;
	mini.y = store.y;
	mini.z = store.z;

	sse::store(maximum,&store);
	maxi.x = store.x;
	maxi.y = store.y;
	maxi.z = store.z;

	int curVertex = 0;
	int curFace   = 0;
	unsigned faceOffset   = 0;
	unsigned vertexOffset = 0;

	transformed		= new Vertex[vertexSum];
	normals			= new Vertex[vertexSum];

	//TODO : This is here bacause the debuging in the visualisation. Change to local variable when no longer needed
	//Triangle *triangles = new Triangle[facesSum];
	triangles = new Triangle[facesSum];

	clock_t start,stop;
	start = clock();

	for(size_t i=0;i<sceneSize;i++) {
		unsigned facesCount  = scene[i].object->facesCount;
		unsigned vertexCount = scene[i].object->vertexCount;

		__m128i vertexSumSse = sse::set(0,vertexOffset,vertexOffset,vertexOffset);

		Vertex*			__restrict vertices	 = scene[i].object->vertices;
		Vertex*			__restrict normalsIn = scene[i].object->normal;
		Triangle*		__restrict faces	 = scene[i].object->faces;
		Transformation* __restrict matrix	 = scene[i].transformation;

		for(unsigned j=0;j<vertexCount;j++) {
			transformed[curVertex].vec = ( vertices[j] * *matrix).vec;
			normals[curVertex++]	   = normalsIn[j];
		};

		__m128i maskFirst    = sse::set(0,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF);

		std::vector<unsigned> &materialIndexes = scene[i].replacementMaterials.size() ? scene[i].replacementMaterials : scene[i].object->materials;

		for(unsigned j=0;j<facesCount;j++) {
			__m128i triangle = sse::load<__m128i>(&faces[j]);						//Load the triangle
			__m128i material = sse::pack(materialIndexes[sse::extract(triangle)]);	//Change the object material number to global material number.

			//Mask the material into the triangle
			triangle = triangle & maskFirst;
			triangle = triangle | material;

			//Convert object triangle indices to global indices
			triangle += vertexSumSse;

			sse::store(triangle,&triangles[j + faceOffset]);
		};

		faceOffset   += facesCount;
		vertexOffset += vertexCount;
	};

	_MM_SET_ROUNDING_MODE(_MM_ROUND_TOWARD_ZERO);

	nodePrimitives = new TailPool<unsigned>(facesSum*2);

	nodes.getNext();
	buildTree<true>(triangles,minimum,maximum,0,0,facesSum,0);
	
	delete nodePrimitives;

	stop  = clock();
//	printf("%f",(double)(stop-start)/CLOCKS_PER_SEC);
}

template<bool initial>
void KdTree::buildTree(Triangle* __restrict faces,__m128 minimum,__m128 maximum,unsigned node,unsigned* indices,unsigned count,unsigned depth) {
	ALIGN16 union {
		uint4    binMin[33];
		unsigned binMinA[33][4];};
	ALIGN16	union {
		uint4    binMax[33];
		unsigned binMaxA[33][4];};

	if(node>39425)
		node=node;

	memset(binMin,0,33*sizeof(uint4));
	memset(binMax,0,33*sizeof(uint4));

	ALIGN16 float4 saveSize;

	__m128 tightMin = minimumInit;
	__m128 tightMax = maximumInit;
	__m128 size		= maximum - minimum;

	//The division of SSE unit has lover precision then x87. SIMD is used in the first part of the algorithm and later
	//the standard floating point is used. Therfore divide here using x87 (Some compilers might actualy use SSE for
	//ALL floating point, but that is OK).
	sse::store(size,&saveSize);
	saveSize.x = 32.0f / saveSize.x;
	saveSize.y = 32.0f / saveSize.y;
	saveSize.z = 32.0f / saveSize.z;

	size = sse::load<__m128>(&saveSize);

	for(unsigned i=0;i<count;i++) {
		if(i==7644)
			i=i;

		//load triangle vertices
		__m128 a,b,c;
		if(initial) {
			a = sse::load<__m128>(&transformed[faces[i].a]);
			b = sse::load<__m128>(&transformed[faces[i].b]);
			c = sse::load<__m128>(&transformed[faces[i].c]);
		} else {
			a = sse::load<__m128>(&transformed[faces[indices[i]].a]);
			b = sse::load<__m128>(&transformed[faces[indices[i]].b]);
			c = sse::load<__m128>(&transformed[faces[indices[i]].c]);
		};

		//Find bounding box of the triangle
		__m128 minExtent = sse::min(sse::min(b,c),a);
		__m128 maxExtent = sse::max(sse::max(b,c),a);

		//Triangled may have bounding box outside node. Clamp triangle extents to node size.
		minExtent = sse::min(minExtent,maximum);
		maxExtent = sse::min(maxExtent,maximum);
		minExtent = sse::max(minExtent,minimum);
		maxExtent = sse::max(maxExtent,minimum);

		tightMin  = sse::min(tightMin,minExtent);
		tightMax  = sse::max(tightMax,maxExtent);

		//Convert extents to bin number
		minExtent = minExtent - minimum;
		maxExtent = maxExtent - minimum;
		
		minExtent = minExtent * size;
		maxExtent = maxExtent * size;

		//Round to integer
		ALIGN16 uint4 rmin,rmax;
		sse::store(sse::f4i(minExtent),&rmin);
		sse::store(sse::f4i(maxExtent),&rmax);

		binMin[rmin.x].x++;
		binMin[rmin.y].y++;
		binMin[rmin.z].z++;

		binMax[rmax.x].x++;
		binMax[rmax.y].y++;
		binMax[rmax.z].z++;
	};

	binMax[31].x += binMax[32].x;
	binMax[31].y += binMax[32].y;
	binMax[31].z += binMax[32].z;

	__m128   minimalCost = minimumInit;
	unsigned minimalIndex[4];

	//Find the bin with best price. This is again done parallel for all axes.
	float sum;
	float xy;
	float xz;
	float yz;
	float xStep;
	float yStep;
	float zStep;
	{
		//The area of the node can be computed incrementaly, instead of recomputation.
		//For example, lets have a node fith x,y,z axes and split is along the x axis.
		//The area of both ends 2*y*z never changes. The area of the rest of sides changes 
		//from bin to bin by value of 2*(x*z+y*z)/binCount. Both values can be predivided by
		//the whole node area. So the area of node at given bin is simple addition and substraction
		ALIGN16 float4 size;
		sse::store(maximum-minimum,&size);
		xy  = 2*size.x*size.y;
		xz  = 2*size.x*size.z;
		yz  = 2*size.y*size.z;
		sum = xy+xz+yz;

		__m128 leftPrice  = sse::set(0.0f,xy/sum,xz/sum,yz/sum);

		float sumNode = sum*32.0f;
		xStep = (xy + xz)/sumNode;
		yStep = (xy + yz)/sumNode;
		zStep = (xz + yz)/sumNode;
		
		__m128 rightPrice = onesInit;
		__m128 priceDiff  = sse::set(0.0f,zStep,yStep,xStep);
		__m128 leftCount  = zeroesInit;
		__m128 rightCount = sse::set4(count);

		for(int i=0;i<31;i++) {
			//Incrementaly compute the are of both childs
			leftPrice  = leftPrice  + priceDiff;
			rightPrice = rightPrice - priceDiff;
	
			//Adjust the object count
			leftCount  = leftCount  + sse::i4f(sse::load<__m128i>(&binMin[i]));
			rightCount = rightCount - sse::i4f(sse::load<__m128i>(&binMax[i]));

			__m128 total  = leftPrice*leftCount + rightPrice*rightCount;
			__m128 mask   = total < minimalCost;

			//Set the best index if some axis has best price
			char xmask = _mm_movemask_ps(mask);
			if(xmask & 0x01) minimalIndex[0] = i;
			if(xmask & 0x02) minimalIndex[1] = i;
			if(xmask & 0x04) minimalIndex[2] = i;
			
			//Combine the current and old best price by masking out the wring values and adding together.
			total		= total & mask;
			mask		= _mm_andnot_ps(mask,minimalCost);
			minimalCost = mask  + total;
		};
	};

	ALIGN16 union {
		float4 cost;
		float  costA[4];
	};
	sse::store(minimalCost,&cost);

	unsigned minIndex = cost.x < cost.y ? 
					    cost.x < cost.z ? 0 : 2 :
						cost.y < cost.z ? 1 : 2;
	unsigned bestIndex = minimalIndex[minIndex];

	//Prefer cutting the empty space at the objects boundaries. This is done rather cumbersomely. Instead checking for
	//empty spaces in the above loop, where it is not that easy, compute the split price at the exact object boundaries
	//in all dimesions here. Find the best price amongst them, and compare to the previous best price.
	__m128 leftRatio  = tightMax - minimum;
	__m128 rightRatio = maximum  - tightMin;
	__m128 nodeSize   = maximum  - minimum;			
	
	//At the tight boundary, one side of split must be empty, so object count in the another must be all triangles in this node.
	__m128 counts = sse::set4(count);
	__m128 init = sse::set(0.0f,xy/sum,xz/sum,yz/sum);
	__m128 step = sse::set(0.0f,(xz + yz)/sum,(xy + yz)/sum,(xy + xz)/sum);
	
	//Compute the price at the tight boundary on both sides of the object.
	leftRatio  = (leftRatio /nodeSize * step + init) * counts;
	rightRatio = (rightRatio/nodeSize * step + init) * counts;

	//Store to ordinary variables and find the best price.
	ALIGN16 float4 left;
	ALIGN16 float4 right;
	sse::store(leftRatio ,&left);
	sse::store(rightRatio,&right);

	unsigned tightIndex;
	bool	 tightLeft;
	
	float tightCost = FLT_MAX;
	if(left.x  < tightCost) {tightCost = left.x; tightIndex=0;tightLeft=true;};
	if(left.y  < tightCost) {tightCost = left.y; tightIndex=1;tightLeft=true;};
	if(left.z  < tightCost) {tightCost = left.z; tightIndex=2;tightLeft=true;};
	if(right.x < tightCost) {tightCost = right.x;tightIndex=0;tightLeft=false;};
	if(right.y < tightCost) {tightCost = right.y;tightIndex=1;tightLeft=false;};
	if(right.z < tightCost) {tightCost = right.z;tightIndex=2;tightLeft=false;};

	ALIGN16 float newMaximum[4];
	ALIGN16 float newMinimum[4];
	ALIGN16 float tightMinR[4];
	ALIGN16 float tightMaxR[4];

	sse::store(maximum,newMaximum);
	sse::store(minimum,newMinimum);
	sse::store(tightMax,tightMaxR);
	sse::store(tightMin,tightMinR);

	//If the best tight boundary is actualy zero volume(it is the side of the node), don't use it
	if(tightLeft)
		if(tightMaxR[tightIndex] >= newMaximum[tightIndex]) tightCost = FLT_MAX;
	else
		if(tightMinR[tightIndex] <= newMinimum[tightIndex]) tightCost = FLT_MAX;

	float boundary;
	float boundaryExtent;
	unsigned leftSum  = 0;
	unsigned rightSum = 0;

	bool addRight = true;

	if(costA[minIndex] < tightCost*0.8f) {
		bestIndex++;
		//Compute the split position based on the bin number. This could be numericaly unstable and object may end
		//up on the different side than they should. This currently doesn't happen on our test scenes. We check for
		//this problem later in code anyway.
		boundary = newMinimum[minIndex] + (float)bestIndex*(newMaximum[minIndex] - newMinimum[minIndex])/32.0f;
		boundaryExtent = bestIndex;

		//We want to allocate memory in large chunks. So find now how many triangles
		//will be on each side of spliting plane. We sum number of triangles starting on left
		//side of the spliting plane and bumber of triangled ending on right sido of the plane
		for(unsigned i = 0;i<bestIndex;i++)
			leftSum  += binMinA[i][minIndex];
		
		for(unsigned i = bestIndex;i<32;i++)
			rightSum += binMaxA[i][minIndex];
	} else {
		//The tight boundary is best
		minIndex = tightIndex;
		if(tightLeft) {
			addRight = false;
			leftSum  = count;
			boundary = tightMaxR[tightIndex];
			boundaryExtent =  128.0f;
		} else {
			rightSum = count;
			boundary = tightMinR[tightIndex];
			boundaryExtent = -128.0f;
		};
	};

	float minClip  = newMinimum[minIndex];
	float maxClip  = newMaximum[minIndex];
	newMaximum[minIndex] = boundary;
	newMinimum[minIndex] = boundary;

	//Now allocate memory. The memory is alocated from the end of the pool and we must return the memory 
	//in the same order as we alocated it. To save memory when building right subtree, alocate memory in this
	//order, so we can free left indexes after left node recursion.
	unsigned* rightFaces = nodePrimitives->getMem(rightSum);
	unsigned* leftFaces  = nodePrimitives->getMem(leftSum);

	unsigned leftIndex  = 0;
	unsigned rightIndex = 0;

	float cellSize = 32.0f/(maxClip-minClip);

	//Now distribute this node triangles to children based on the spliting plane possition. This might suffer from the
	//above mentioned instability. Solution might be to compute the bin possition again, and compare that instead of
	//comparing the extends.
	for(unsigned i=0;i<count;i++) {
		//load triangle vertices
		float a,b,c;
		if(initial) {
			a = *(&transformed[faces[i].a].x + minIndex);
			b = *(&transformed[faces[i].b].x + minIndex);
			c = *(&transformed[faces[i].c].x + minIndex);

			//Find bounding box of the triangle
			float minExtent = min(a,min(b,c));
			float maxExtent = max(a,max(b,c));

			if(minExtent < boundary)
				leftFaces[leftIndex++] = i;
					
			if(maxExtent > boundary) 
				rightFaces[rightIndex++] = i;
		} else {
			a = *(&transformed[faces[indices[i]].a].x + minIndex);
			b = *(&transformed[faces[indices[i]].b].x + minIndex);
			c = *(&transformed[faces[indices[i]].c].x + minIndex);

			//Find bounding box of the triangle
			float minExtent = (min(a,min(b,c)) - minClip)*cellSize;
			float maxExtent = (max(a,max(b,c)) - minClip)*cellSize;

			if(minExtent < boundaryExtent)
				leftFaces[leftIndex++] = indices[i];
					
			if(maxExtent > boundaryExtent) 
				rightFaces[rightIndex++] = indices[i];
		};
	};

	if(leftIndex > leftSum || rightIndex > rightSum) {
		throw 1;
	};

	//Allocate nodes structures
	unsigned leftChild = nodes.getNext();
	KdNode *nodePtr = nodes[node];
	nodePtr->split = boundary;
	nodePtr->flags = minIndex;

	__m128 maximumLeft,minimumRight;
	maximumLeft  = sse::load<__m128>(newMaximum);
	minimumRight = sse::load<__m128>(newMinimum);

	//Build left subtree
	if(leftIndex <= 128)
		buildNodeExact  (faces,minimum,maximumLeft,leftChild,leftFaces,leftIndex,depth+1);
	else 
		buildTree<false>(faces,minimum,maximumLeft,leftChild,leftFaces,leftIndex,depth+1);
		

	//The nodes have implicit pointers to left child. We must allocate right node after left node recursion
	//to garantee all implicit pointed node to be finished.
	unsigned rightChild  = nodes.getNext();
	nodes[node]->setChild(rightChild);
	nodePrimitives->returnMem(leftSum);

	//Build right subtree
	if(rightIndex <= 128)
		buildNodeExact  (faces,minimumRight,maximum,rightChild,rightFaces,rightIndex,depth+1);
	else
		buildTree<false>(faces,minimumRight,maximum,rightChild,rightFaces,rightIndex,depth+1);
		

	nodePrimitives->returnMem(rightSum);
};


//Distribute the triangles to the children node.
template<bool addLeft>
void distribute(Vertex* transformed,Triangle* __restrict faces,unsigned* indices,unsigned* leftFaces,unsigned* rightFaces,float boundary,float minimum,float maximum,unsigned minIndex,unsigned count,unsigned &leftOut,unsigned &rightOut) {
	unsigned leftIndex=0,rightIndex=0;

	for(unsigned i=0;i<count;i++) {
		//load triangle vertices
		float a,b,c;
		a = *(&transformed[faces[indices[i]].a].x + minIndex);
		b = *(&transformed[faces[indices[i]].b].x + minIndex);
		c = *(&transformed[faces[indices[i]].c].x + minIndex);

		//Find extents of the triangle
		float minExtent = min(min(b,c),a);
		float maxExtent = max(max(b,c),a);
		minExtent = max(minExtent,minimum);
		maxExtent = min(maxExtent,maximum);

		if(minExtent < boundary)
			leftFaces[leftIndex++]   = indices[i];
   		
		if(maxExtent > boundary)
			rightFaces[rightIndex++] = indices[i];

		//If the extents are the same, the triangle is coplanar with the spliting plane.
		//Add the triangle to the appropiate side, based on the parameter.
		if(minExtent == boundary && maxExtent==boundary)
			if(addLeft)
				leftFaces[leftIndex++] = indices[i];
			else
				rightFaces[rightIndex++] = indices[i];
	};

	leftOut	 = leftIndex;
	rightOut = rightIndex;
};

inline void KdTree::addEvent(Event* events,float minimum,float maximum,unsigned &count) {
	if(minimum == maximum) {	// The extents are the same. The triangle is coplanar and generates only one event.
		events[count  ].pos  = minimum;
		events[count++].type = INPLANE;
	} else {
		events[count  ].pos  = minimum;
		events[count++].type = STARTS;
		events[count  ].pos  = maximum;
		events[count++].type = ENDS;
	};
};	

void KdTree::bestPriceExact(Event *events,float minimum,float maximum,float step,float init,unsigned Eventcount,unsigned count,SplitInfo &info) {
	unsigned leftObjects  = 0;
	unsigned planeObjects = 0;
	unsigned rightObjects = count;

	float size  = maximum - minimum;
	float best  = FLT_MAX;
	float split;
	unsigned bestLeft;
	unsigned bestRight;
	unsigned addLeft;

	if(size == 0) {
		info.price = FLT_MAX;
		info.split = 0.0f;
		info.leftCount  = 0;
		info.rightCount = 0;
		info.addLeft	= 0;
		return;
	};

	//The best spliting position search algorithm as decribed in 
	//"On building fast kd-Trees for Ray Tracing, and on doing that in O(N log N)"

	unsigned i=0;
	while (i<Eventcount) {
		unsigned starts  = 0;
		unsigned ends    = 0;
		unsigned inplane = 0;

		float boundary = events[i].pos;
		
		while(i<Eventcount && events[i].pos == boundary && events[i].type == ENDS   ) {i++; ends++;};
		while(i<Eventcount && events[i].pos == boundary && events[i].type == INPLANE) {i++; inplane++;};
		while(i<Eventcount && events[i].pos == boundary && events[i].type == STARTS ) {i++; starts++;};

		rightObjects-=ends;
		rightObjects-=inplane;

		float surface		= (boundary - minimum)/size;
		float leftSurface	= surface*step+init;
		float rightSurface	= (1.0f-surface)*step+init;
		
		float priceLeft  = 1.0f + leftSurface*(leftObjects+inplane) + rightSurface*rightObjects;
		float priceRight = 1.0f + leftSurface*leftObjects + rightSurface*(rightObjects+inplane);

		/*if(leftObjects  == 0 || rightObjects+inplane == 0) priceRight*=0.8f;
		if(rightObjects == 0 || leftObjects+inplane == 0) priceLeft  *=0.8f;*/

		if(priceLeft < priceRight) {
			if(priceLeft < best && boundary!=minimum && boundary!=maximum) {
				best	  = priceLeft;
				split	  = boundary;
				addLeft	  = true;
				bestLeft  = leftObjects+inplane;
				bestRight = rightObjects;
			};
		} else {
			if(priceRight < best && boundary!=minimum && boundary!=maximum) {
				best	  = priceRight;
				split	  = boundary;
				addLeft	  = false;
				bestLeft  = leftObjects;
				bestRight = rightObjects+inplane;
			};
		};

		leftObjects+=starts;
		leftObjects+=inplane;
	};

	info.price = best;
	if(best != FLT_MAX) {
		info.split = split;
		info.leftCount  = bestLeft;
		info.rightCount = bestRight;
		info.addLeft	= addLeft;
	};
};

void KdTree::buildNodeExact(Triangle* __restrict faces,__m128 minimum,__m128 maximum,unsigned node,unsigned* indices,unsigned count,unsigned depth) {
	if(node==8332)
		node=node;

	if(count <= 2) {
		KdNode *nodePtr = nodes[node];
		nodePtr->flags  = LEAF;
		nodePtr->count |= count<<2;

		if(count) {
			unsigned triangleStart = nodeTriangles.getMore(count);
			nodePtr->faces = triangleStart;
			Triangle* start = nodeTriangles[triangleStart];

			for(unsigned i=0;i<count;i++)
				start[i] = faces[indices[i]];
		};

		return;
	};

	if(count >= 256)
		throw 1;

	static Event xEvents[512];
	static Event yEvents[512];
	static Event zEvents[512];
	unsigned xCount = 0;
	unsigned yCount = 0;
	unsigned zCount = 0;

	for(unsigned i=0;i<count;i++) {
		//load triangle vertices
		__m128 a = sse::load<__m128>(&transformed[faces[indices[i]].a]);
		__m128 b = sse::load<__m128>(&transformed[faces[indices[i]].b]);
		__m128 c = sse::load<__m128>(&transformed[faces[indices[i]].c]);

		//Find bounding box of the triangle
		__m128 minExtent = sse::min(sse::min(b,c),a);
		__m128 maxExtent = sse::max(sse::max(b,c),a);

		//Triangled may have bounding box outside node. Clamp triangle extents to node size.
		minExtent = sse::min(minExtent,maximum);
		maxExtent = sse::min(maxExtent,maximum);
		minExtent = sse::max(minExtent,minimum);
		maxExtent = sse::max(maxExtent,minimum);

		ALIGN16 float4 rmin,rmax;
		sse::store(minExtent,&rmin);
		sse::store(maxExtent,&rmax);

		addEvent(xEvents,rmin.x,rmax.x,xCount);
		addEvent(yEvents,rmin.y,rmax.y,yCount);
		addEvent(zEvents,rmin.z,rmax.z,zCount);
	};

	std::sort(PlainIterator<Event>(xEvents),PlainIterator<Event>(xEvents + xCount));
	std::sort(PlainIterator<Event>(yEvents),PlainIterator<Event>(yEvents + yCount));
	std::sort(PlainIterator<Event>(zEvents),PlainIterator<Event>(zEvents + zCount));

	SplitInfo infoX,infoY,infoZ;
	SplitInfo *best;

	//Prepare bounding boxes of child nodes.
	ALIGN16 float newMaximum[4];
	ALIGN16 float newMinimum[4];
	
	sse::store(maximum,newMaximum);
	sse::store(minimum,newMinimum);

	if(node > 39425)
		node = node;

	{
	ALIGN16 float4 size;
	sse::store(maximum-minimum,&size);
	float xy  = 2*size.x*size.y;
	float xz  = 2*size.x*size.z;
	float yz  = 2*size.y*size.z;
	float sum = xy+xz+yz;

	bestPriceExact(xEvents,newMinimum[0],newMaximum[0],(xy + xz)/sum,yz/sum,xCount,count,infoX);
	bestPriceExact(yEvents,newMinimum[1],newMaximum[1],(xy + yz)/sum,xz/sum,yCount,count,infoY);
	bestPriceExact(zEvents,newMinimum[2],newMaximum[2],(xz + yz)/sum,xy/sum,zCount,count,infoZ);
	}

	unsigned minIndex;

	if(infoX.price < infoY.price) {
		if(infoX.price < infoZ.price) {
			minIndex = 0;
			best = &infoX;
		} else {
			minIndex = 2;
			best = &infoZ;
		};
	} else {
		if(infoY.price < infoZ.price) {
			minIndex = 1;
			best = &infoY;
		} else {
			minIndex = 2;
			best = &infoZ;
		};
	};

	//if(best->price > count*0.9f) {
	if(best->price > count) {	//Create leaf node if the price is not decreased enough by spliting.
		KdNode *nodePtr = nodes[node];
		nodePtr->flags  = LEAF;
		nodePtr->count |= count<<2;

		if(count) {
			unsigned triangleStart = nodeTriangles.getMore(count);
			nodePtr->faces = triangleStart;
			Triangle* start = nodeTriangles[triangleStart];

			for(unsigned i=0;i<count;i++)
				start[i] = faces[indices[i]];
		};

		return;
	};

	float boundary = best->split;
	//Now allocate memory. The memory is alocated from the end of the pool and we must return the memory 
	//in the same order as we alocated it. To save memory when building right subtree, alocate memory in this
	//order, so we can free left indexes after left node recursion.
	unsigned* rightFaces = nodePrimitives->getMem(best->rightCount);
	unsigned* leftFaces  = nodePrimitives->getMem(best->leftCount);

	unsigned leftIndex  = 0;
	unsigned rightIndex = 0;

	if(best->addLeft)
		distribute<true>(transformed,faces,indices,leftFaces,rightFaces,boundary,newMinimum[minIndex],newMaximum[minIndex],minIndex,count,leftIndex,rightIndex);
	else 
		distribute<false>(transformed,faces,indices,leftFaces,rightFaces,boundary,newMinimum[minIndex],newMaximum[minIndex],minIndex,count,leftIndex,rightIndex);

	//assert(!(leftIndex != best->leftCount || rightIndex != best->rightCount));
	if(leftIndex != best->leftCount || rightIndex != best->rightCount) {
		leftIndex = leftIndex;
		throw 1;
	};

	//if(boundary==newMinimum[minIndex] )	    boundary+=boundary*0.01f;
	//else if(boundary==newMaximum[minIndex]) boundary-=boundary*0.01f;

	newMaximum[minIndex] = boundary;
	newMinimum[minIndex] = boundary;

	//Allocate nodes structures
	unsigned leftChild = nodes.getNext();
	KdNode*  nodePtr   = nodes[node];
	nodePtr->split = boundary;
	nodePtr->flags = minIndex;

	__m128 maximumLeft,minimumRight;
	maximumLeft  = sse::load<__m128>(newMaximum);
	minimumRight = sse::load<__m128>(newMinimum);

	//Build left subtree
	buildNodeExact(faces,minimum,maximumLeft,leftChild,leftFaces,best->leftCount,depth+1);
	nodePrimitives->returnMem(best->leftCount);

	//The nodes have implicit pointers to left child. We must allocate right node after left node recursion
	//to garantee all implicit pointed node to be finished.
	unsigned rightChild  = nodes.getNext();
	nodes[node]->setChild(rightChild);

	//Build right subtree
	buildNodeExact(faces,minimumRight,maximum,rightChild,rightFaces,best->rightCount,depth+1);
	nodePrimitives->returnMem(best->rightCount);
};

KdTree::~KdTree(void) {
	delete[] transformed;
	delete[] triangles;
	delete[] normals;

	if(videoAllocated) {
		cudaFree(savedTriangles);
		cudaFreeArray(savedTree);
		cudaFreeArray(savedVertices);
		cudaFreeArray(savedNormals);
	};
}


bool KdTree::prepareCudaTexture(float3 &minimum, float3 &maximum,
						    cudaArray* &cudaTree,uint4* &triangles,
							cudaArray* &vertices,cudaArray* &normalsOut) {
	minimum = mini;
	maximum = maxi;

	cudaChannelFormatDesc channelDesc;
	cudaExtent			  gridSize;

	cudaTree	= 0;
	vertices	= 0;
	normalsOut	= 0;
	triangles	= 0;

	try {

	unsigned sideX = 256;
	unsigned sideY = (unsigned)nodes.size()/sideX + 1;

	channelDesc = cudaCreateChannelDesc<uint2>();
	gridSize	= make_cudaExtent(sideX, sideY, 0);

	SAFE_ALLOC(cudaMallocArray(&cudaTree, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(cudaTree,0,0,nodes[0],nodes.size()*sizeof(uint2),cudaMemcpyHostToDevice);


/////////////////////////////

	sideX = 256;
	sideY = vertexSum/sideX + 1;

	channelDesc = cudaCreateChannelDesc<float4>();
	gridSize	= make_cudaExtent(sideX, sideY, 0);

	SAFE_ALLOC(cudaMallocArray(&vertices, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(vertices,0,0,transformed,vertexSum*sizeof(float4),cudaMemcpyHostToDevice);

	SAFE_ALLOC(cudaMallocArray(&normalsOut, &channelDesc, sideX,sideY));
	cudaMemcpyToArray(normalsOut,0,0,normals,vertexSum*sizeof(float4),cudaMemcpyHostToDevice);

/////////////////////////////
	
	SAFE_ALLOC(cudaMalloc((void**)&triangles,nodeTriangles.size()*sizeof(Triangle)));
	cudaMemcpy(triangles,nodeTriangles[0],nodeTriangles.size()*sizeof(Triangle),cudaMemcpyHostToDevice);

	} catch(OutOfMemory) {
		if(triangles)  cudaFree(triangles);
		if(cudaTree)   cudaFreeArray(cudaTree);
		if(vertices)   cudaFreeArray(vertices);
		if(normalsOut) cudaFreeArray(normalsOut);
		throw OutOfMemory();
	};

	videoAllocated  = true;
	savedTree		= cudaTree;
	savedTriangles  = triangles;
	savedVertices	= vertices;
	savedNormals	= normalsOut;

	return true;
};