/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#include <vector_types.h>
#include <cuda_runtime_api.h>

#include "Utils/Memory.h"

#include "Object.h"
#include "Transformations/SceneNode.h"

const unsigned X_AXIS = 0x00;
const unsigned Y_AXIS = 0x01;
const unsigned Z_AXIS = 0x02;
const unsigned LEAF   = 0x03;

struct KdNode {
	union {
		unsigned	flags;
		unsigned	rightChild;
		unsigned	count;
	};

	union {
		float		split;
		unsigned	faces;		
	};

	void setChild(unsigned child) {rightChild |= child<<2;};
};

//! Kd tree builder
class KdTree {
	//! The type of event in exact SAH build
	/**
	*	The element must be in this order, as thir integer values are used for sorting the events.
	*/
	enum EventType {
		ENDS,		//!< The triangle end at this position.
		INPLANE,	//!< The triangle is coplanar with the spliting plane
		STARTS		//!< The triangle starts at this position.
	};

	//! Triangle extents used for sorting during exact SAH.
	struct Event {
		float	  pos;		//!< The possition of the event.
		EventType type;		//!< The type of event.

		//! Define ordering of events.
		/**
		*	The events positions are compared first. If they are the same, decide using the type of the event.
		*/
		bool operator<(const Event &right) {	
			if(pos<right.pos)  return true;
			if(pos==right.pos) return type < right.type;
			return false;
		};
	};

	//! Used to return spliting plane info from exact search.
	struct SplitInfo {
		float			price;		//!< The price at the spliting position.
		float			split;		//!< The position of the spliting plane.
		unsigned short	leftCount;	//!< Number of triangles left to the plane.
		unsigned short	rightCount; //!< Number of triangles right to the plane.
		unsigned		addLeft;	//!< On witch side should we add coplanar triangles.
	};

	//! Add new event to the events array
	/**
	*	@param[in,out] events	The array of events.
	*	@param		   minimum	The minimum extent of triangle for this event.
	*	@param		   maximum	The maximum extent of triangle for this event.
	*	@param[in,out] count		The current number of events. Return new number of events.
	*/
	inline void addEvent(Event* events,float minimum,float maximum,unsigned &count);

public:
	float3 mini;	//!< Minimum extent of the root node, prepared for CUDA.
	float3 maxi;	//!< Maximum extent of the root node, prepared for CUDA.

	int vertexSum;	//!< Total number of vertices of objects included in the grid.
	int facesSum;	//!< Total number of faces of objects included in the grid.

	Vertex		*transformed;	//!< Vertices of objects in grid, transformed into world space.
	Vertex		*normals;		//!< Normals associated with vertices. Has one normal for each vertex.
	Triangle    *triangles;		//!< Temporal triangle indexes collected from all scene objects.
								/*!< \warning This is here for debuging and could be removed in later versions.*/

	GrowArray<KdNode>   nodes;				//!< Nodes of the tree	
	GrowArray<Triangle> nodeTriangles;		//!< Faces of the tree. 
											/*!< Referenced from tree nodes. Has size of total number of triangles referenced in leaf nodes. This can be several time the number in totalPoly. */
	TailPool<unsigned> *nodePrimitives;		//!< Temporal indicies used to pass relevant faces to children nodes.

	bool		videoAllocated;	//!< True if video memory is allocated.
	cudaArray*	savedTree;		//!< Saved pointer for video memory dealocation.
	uint4*	    savedTriangles; //!< Saved pointer for video memory dealocation.
	cudaArray*  savedVertices;  //!< Saved pointer for video memory dealocation.
	cudaArray*  savedNormals;   //!< Saved pointer for video memory dealocation.

	//! Find best spliting position 
	/**
	*	@param[in] events		Store the triangles extent.
	*	@param	   minimum		The minimum extent of current node.
	*	@param	   size			The size of current node.
	*	@param	   step			The area of node sides that <b>are</b> cut by spliting plane.
	*	@param	   init			The area of node sides that <b>are not</b> cut by spliting plane.
	*	@param	   Eventcount	The number of events.
	*	@param	   count		The number of triangles. Must by provided separately because some events might be 
	*							inplane events. Therefore one cannot be deduced simply from the other.
	*	@param[out] info		Return the infor about spliting plane.		
	*/
	void bestPriceExact(Event *events,float minimum,float maximum,float step,float init,unsigned Eventcount,unsigned count,SplitInfo &info);

	//! Build node using aproximate min-max binning
	/**
	*	@tparam initial True if building the root node. In that case, the indices are ignored and the faces are accessed directly.
	*	@param faces	The source triangles colectected from all scene objects.
	*	@param minimum	The minimum extent of the new node.
	*	@param maximum  The maximum extent of the new node.
	*	@param node		The node number assigned to the new node.
	*   @param indices	The indices to the triangles witch belongs to this node
	*	@param count	The number of triangles for this node
	*	@param depth	The depth of the cell in the tree
	*/
	template<bool initial>
	void buildTree(Triangle* __restrict faces,__m128 minimum,__m128 maximum,unsigned node,unsigned* indices,unsigned count,unsigned depth);

	//! Build node using exact SAH.
	/**
	*	@param faces	The source triangles colectected from all scene objects.
	*	@param minimum	The minimum extent of the new node.
	*	@param maximum  The maximum extent of the new node.
	*	@param node		The node number assigned to the new node.
	*   @param indices	The indices to the triangles witch belongs to this node
	*	@param count	The number of triangles for this node
	*	@param depth	The depth of the cell in the tree
	*/
	void buildNodeExact(Triangle* __restrict faces,__m128 minimum,__m128 maximum,unsigned node,unsigned* indices,unsigned count,unsigned depth);

public:
	//! Build new tree for scene
	/**
	*	@param scene The scene for whitch the tree should be build.
	*/
	KdTree(Scene &scene);

	//! Free host and if necesary video card memory
	~KdTree(void);

	//! Prepare tree for use in CUDA
	/*!
	*	Copy the tree data to video card's linear memory. Uses texture to store the 
	*	nodes,vertices,normals and linear memory to store triangles. 
	*
	*   @param minimum		Returns minimum extent of the root node.
	*	@param maximum		Returns maximum extent of the root node.
	*	@param cudaTree		Pointer to tree nodes in video memory
	*	@param triangles	Pointer to triangles in video memory
	*	@param vertices		Pointer to vertices in video memory
	*	@param normalsOut   Pointer to normals in video memory
	*	@return				true if succesfull, false if out of memory
	*/
	bool prepareCudaTexture(float3 &minimum, float3 &maximum,
						    cudaArray* &cudaTree,uint4* &triangles,
							cudaArray* &vertices,cudaArray* &normalsOut);
};
