/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <vector_types.h>
#include <cuda_runtime_api.h>

#include "Object.h"
#include "Transformations/SceneNode.h"

struct Cell {
	unsigned int offset;
	unsigned int count;
};


//! Uniform grid builder
class UniformGrid {
public:
	int dimX;			//!< Number of cells in x dimension.
	int dimY;			//!< Number of cells in y dimension.
	int dimZ;			//!< Number of cells in z dimension.

	float  cuCellSize;	//!< Size of one cell in world space. The size is the same in all dimensions.
	float3 cuMinimum;	//!< Minimum extent of the grid, prepared for CUDA.
	float3 cuMaximum;   //!< Maximum extent of the grid, prepared for CUDA.

	unsigned	 totalPoly;		//!< Total number of faces of objects included in the grid.
	unsigned	 totalVertex;	//!< Total number of vertices of objects included in the grid.

	Cell		*grid;			//!< Cells of the grid.	

	Vertex		*transformed;	//!< Vertices of objects in grid, transformed into world space.
	Vertex		*normals;		//!< Normals associated with vertices. Has one normal for each vertex.
	Triangle	*cellTriangles; //!< Faces of the grid. 
								/*!< Referenced from grid cells. Has size of total number of triangles referenced in cells. This can be several time the number in totalPoly. */
							
	bool		videoAllocated; //!< True if video memory is allocated.
	cudaArray*	savedGrid;		//!< Saved pointer for video memory dealocation.
	uint4*	    savedTriangles;	//!< Saved pointer for video memory dealocation.
	cudaArray*  savedVertices;	//!< Saved pointer for video memory dealocation.
	cudaArray*  savedNormals;	//!< Saved pointer for video memory dealocation.

public:
	UniformGrid(Scene &scene,int minimalSideSize = -1);
	~UniformGrid(void);

	//! Prepare grid for use in CUDA
	/*!
	*	Copy the grid data to video card's linear memory. Uses texture to store the 
	*	cells,vertices,normals and linear memory to store triangles. 
	*
	*   @param minimum		Returns minimum extent of the grid.
	*	@param maximum		Returns maximum extent of the grid.
	*	@param cellSize		Returns size of the cell
	*   @param dim			Returns size of grid
	*	@param cudaGrid		Pointer to cells in video memory
	*	@param triangles	Pointer to triangles in video memory
	*	@param vertices		Pointer to vertices in video memory
	*	@param normalsOut   Pointer to normals in video memory
	*	@return				true if succesfull, false if out of memory
	*/
	bool prepareCudaTexture(float3 &minimum, float3 &maximum, float &cellSize, uint3 &dim,
					 cudaArray* &cudaGrid,uint4* &triangles,cudaArray* &vertices,cudaArray* &normalsOut);

};
