/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <vector>
#include <map>
#include <string>

#include "Transformations/Transformation.h"
#include "FormatImport/FileLoader.h"
#include "Object.h"


struct ObjectInstance {
	Transformation		  *transformation;
	std::vector<unsigned>  replacementMaterials;
	Object				  *object;

	ObjectInstance(Transformation *transformation,Object *object) : transformation(transformation),object(object) {};
};

class Scene:public std::vector<ObjectInstance>  {
public:
	~Scene() {
		for(unsigned i=0;i<size();i++) {
			delete at(i).transformation;
			delete at(i).object;
		};
		MaterialManager::instance()->reset();
	};
};


//! Global singleton class used as a wrapper around file loader classes. 
/**
*	All file loaders must register them selves with this class. As it is singleton it can be succesfully
*   complted during global variable initialization. Assigning the addLoader method to a dummy global variable
*	results in automaticaly registering file loader at program initialization time. The is singleton correctly 
*	destroyed at program exit, so all registered file loader classes destructors are called.
*/
class SceneLoader {
	static SceneLoader* objInstance;

	std::map<std::string,FileLoader*>	loaderMap;	//!< Maps the file extension to file loader.

	SceneLoader();
	~SceneLoader();
public:

	//! Get the singleton instance. 
	/**
	*	Creates new instance if not already initialized.
	*	@return The SceneLoader singelton instance.
	*/
	static SceneLoader*	instance() {
		if(!objInstance) {
			//Placing static instance into the motod result in the object destruction at program exit.
			//If placed as static class member, that would not happen
			static SceneLoader sceneLoader;	
			objInstance = &sceneLoader;
		};
		return objInstance;
	};

	//! Register new file loader.
	/**
	*	Creates new instance if not already initialized. 
	*	@return Nothing. This is here, so it can be used as dummy global variable initializer.
	*/
	unsigned addLoader(FileLoader* fileLoader,std::string name);

	//! Load file and return it as a scene.
	/**
	*	The object in the file is loaded and returned as part of the scene. If file contains no materials
	*	the default material is inserted. The center of object can be optionaly moved to cooridinates origin.
	*	
	*	@param	filename	Path to the file containing the scene.
	*	@param	recenter	Move the center of object to origin.
	*	@return The loaded scene
	*/
	Scene*	 loadScene(const char* filename,bool recenter = false);


	//! Load file and return as a simple object.
	/**
	*	@param	filename	Path to the file containing the scene.
	*/
	Object*	 load(const char* filename);
};