/*----------------------------------------------------------------------
  File:    animation.c
  Purpose: functions and data structures for animation
           This file contains all functions you need to get the animted
           transform out of an struct AnimationList
----------------------------------------------------------------------*/

#include <math.h>
#include <string.h>
#include "animation.h"
#include "quat.h"

#ifdef WIN32
#include <float.h>
#endif

/*----------------------------------------------------------------------
  _GetTranslation()
  Purpose: same as GetTranslation, but you give the Animation *a, as an
  in-argument instead of the name of the Animation.
----------------------------------------------------------------------*/
bool _GetTranslation(Animation* animation, double time, double Tout[3])
{
   Point3 point;
   void* p =animation->translations;
   if(p)
   {
      KB_PosInterpolate0( p, time, &point);
      Tout[0]=point.x;
      Tout[1]=point.y;
      Tout[2]=point.z;
      return true;
   }
   Tout[0]=Tout[1]=Tout[2]=0.0;   
   return false;
}

/*----------------------------------------------------------------------
  _GetRotation()
  Purpose: same as GetRotation, but you give the Animation *a, as an
  in-argument instead of the name of the Animation.
----------------------------------------------------------------------*/
bool _GetRotation(Animation* animation, double time, double Rout[4]) {
  Rotation3 rotation;
  void* r = animation->rotations;
  if(r) {
     KB_RotInterpolate( r, time, &rotation);
     Rout[0]=rotation.x;
     Rout[1]=rotation.y;
     Rout[2]=rotation.z;
     Rout[3]=rotation.angle;
     return true;
  }
  Rout[0]=1.0; Rout[1]=0.0; Rout[2]=0.0; Rout[3]=0.0;
  return false;
}

/*----------------------------------------------------------------------
  _GetScale()
  Purpose: same as GetScale, but you give the Animation *a, as an
  in-argument instead of the name of the Animation.
----------------------------------------------------------------------*/
bool _GetScale(Animation* animation, double time, double Sout[3]) {
   Point3 scale;
   void* s = animation->scales;
   if(s)
   {
      KB_PosInterpolate0(s, time, &scale); 
      Sout[0]=scale.x;
      Sout[1]=scale.y;
      Sout[2]=scale.z;
      return true;
   }
   Sout[0]=1.0; Sout[1]=1.0; Sout[2]=1.0;
   return false;
}

/*----------------------------------------------------------------------
  _GetMatrix()
  Purpose: same as GetMatrix, but you give the Animation *a, as an
  in-argument instead of the name of the Animation. Another exception
  is that there is no return value, instead the matrix is the identity
  if there were no animation.
----------------------------------------------------------------------*/
void _GetMatrix(Animation* animation, double time, double m[4][4])
{
  /* m = translation * rotation * scale */
  double drot[4];
  Rotation3 RR;
  Quaternion q;
  Animation* a = (Animation*)animation;
  
  /* rotation */
  if(a->rotations)
  {
    _GetRotation(a, time, drot);
    RR.x=drot[0];
    RR.y=drot[1];
    RR.z=drot[2];
    RR.angle=drot[3];
    FromAngleAxis(&q, RR.angle, RR.x, RR.y, RR.z);
    QuatToMatrix(&q, m); // Also sets bottom row to 0 0 0 1 
  }
  else
  {
    m[0][0] = 1; m[0][1] = 0; m[0][2] = 0; m[0][3] = 0;
    m[1][0] = 0; m[1][1] = 1; m[1][2] = 0; m[1][3] = 0;
    m[2][0] = 0; m[2][1] = 0; m[2][2] = 1; m[2][3] = 0;
    m[3][0] = 0; m[3][1] = 0; m[3][2] = 0; m[3][3] = 1;
  }

  // concatenate with scaling
  if(a->scales)
  {
     double S[3];
     _GetScale(a, time, S);
     m[0][0] *= S[0]; m[0][1] *= S[1]; m[0][2] *= S[2];
     m[1][0] *= S[0]; m[1][1] *= S[1]; m[1][2] *= S[2];
     m[2][0] *= S[0]; m[2][1] *= S[1]; m[2][2] *= S[2];
  }

  /* translation */
  if(a->translations)  
  {
     double P[3];
     _GetTranslation(a, time, P);
     m[0][3] = P[0];
     m[1][3] = P[1];
     m[2][3] = P[2];
  }
}

/*----------------------------------------------------------------------
  Note: A transformation is a concatenation of a scaling matrix S, 
  and a rotation matrix R, and a translation matrix T. The order
  is: T*R*S  i.e., the scaling should be done first, then the rotation,
  and finally the translation.
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  GetCamera()
  Purpose: to get the position, the view direction and the up vector
     of the camera if it is animated. The name of the animation for
     the camera is "camera".
  In parameters:
    al   - the linked list of Animations
    time - the time of the wanted transformation
  Out parameters:
    gotPosition - by the end of the call, this will be 1 if there was a
                  position for the camera, else 0
    viewPos - the camera position (only set if gotPosition returns 1) 
    gotDirection - by the end of the call, this will be 1 if there was a
                  view direction and up vector for the camera, else 0
    viewDir - the camera direction (only set if gotDirection returns 1) 
    viewUp - the camera up vector (only set if gotDirection returns 1) 
  Returns:
    void
----------------------------------------------------------------------*/
/*void GetCamera(struct AnimationList *al,double time,
	       int *gotPosition,double viewPos[3],
	       int *gotDirection, double viewDir[3], double viewUp[3]) {
	Animation *a;
	// is there an animation called "camera", then use it to animate the camera
	a=FindAnimation("camera",al);
	if(a) {
		double m[4][4];
      
		// get position of the camera, if any
		if(a->translations) {
			_GetTranslation(a,time,viewPos);
			*gotPosition=1;
		} else
			*gotPosition=0;

		// get view direction and up vector, if any
		if(a->rotations) {
			// get matrix (we use only the upper 3x3 matrix to extract dir, and up
			_GetMatrix(a,time,m);
		 
			// the viewer looks along the negative z-axis, with the y-axis up
			viewDir[0]=-m[0][2];
			viewDir[1]=-m[1][2];
			viewDir[2]=-m[2][2];
		 
			// the y-axis is up
			viewUp[0]=m[0][1];
			viewUp[1]=m[1][1];
			viewUp[2]=m[2][1];
		 
			*gotDirection=1;
		} else
			*gotDirection=0;
   } else {
		// we did not find any position nor a direction
		*gotPosition=0;
		*gotDirection=0;
   }
}*/

