/*----------------------------------------------------------------------
  Previewer for Benchmark for Animated Ray Tracings: BART
  File:    parse.c
           Based on readnff.c (part of SPD by Eric Haines) by Eduard Schwan
  Purpose: simple parser for data files, includeing args parsing

----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "parse.h"
#include "animation.h"

#include <map>
#include <string>

#include <vector_functions.h>
#include <vector_types.h>

using std::map;
using std::string;

#ifndef  M_PI
#define  M_PI 3.1415926535897932384626433
#endif

//#include "Representation.h"

#define X 0 
#define Y 1  
#define Z 2  
#define W 3  
             
#define R 0  
#define G 1  
#define B 2  
#define A 3 /* alpha */

/*----------------------------------------------------------------------
  parseComment()
  Description:
    "#" [ string ]
    or 
    "%" [ string ]

  Format:
    # [ string ]
    or
    % [ string ]

    As soon as a "#" (or "%") character is detected, the rest of the line is
    considered a comment. 
----------------------------------------------------------------------*/
static void parseComment(FILE *f) {
   char str[1000];
   fgets(str, 1000, f);
}

/*----------------------------------------------------------------------
  parseViewpoint()
  Description:
    "v"
    "from" Fx Fy Fz
    "at" Ax Ay Az
    "up" Ux Uy Uz
    "angle" angle
    "hither" hither
    "resolution" xres yres

  Format:

    v
    from %g %g %g
    at %g %g %g
    up %g %g %g
    angle %g
    hither %g
    resolution %d %d

  The parameters are:

    From:  the eye location in XYZ.
    At:    a position to be at the center of the image, in XYZ world
	   coordinates.  A.k.a. "lookat".
    Up:    a vector defining which direction is up, as an XYZ vector.
    Angle: in degrees, defined as from the center of top pixel row to
	   bottom pixel row and left column to right column.
           In AFF, if the width is different from the height, then
           we interpret the Angle as the FOV in the y-direction (from top to bottom),
           and we set the aspect ratio to width/height.
    Resolution: in pixels, in x and in y.

  Note that no assumptions are made about normalizing the data (e.g. the
  from-at distance does not have to be 1).  Also, vectors are not
  required to be perpendicular to each other.

  For all databases some viewing parameters are always the same:
    Yon is "at infinity."
    Aspect ratio is 1.0.

  A view entity must be defined before any objects are defined (this
  requirement is so that NFF files can be used by hidden surface machines).
----------------------------------------------------------------------*/
static void parseViewpoint(FILE *fp) {
	float3 from;
	float3 at;
	float3 up;
	float fov_angle;
	float hither;
	int resx;
	int resy;

	if(fscanf(fp, " from %f %f %f", &from.x, &from.y, &from.z) != 3)
		goto fmterr;

	if(fscanf(fp, " at %f %f %f", &at.x, &at.y, &at.z) != 3)
		goto fmterr;

	if(fscanf(fp, " up %f %f %f", &up.x, &up.y, &up.z) != 3)
		goto fmterr;

	if(fscanf(fp, " angle %f", &fov_angle) != 1)
		goto fmterr;

	if(fscanf(fp, " hither %f", &hither) !=1)
		goto fmterr;
	if(hither<0.0001) hither=1.0f;
	  
	if(fscanf(fp, " resolution %d %d", &resx, &resy) !=2)
		goto fmterr;

	// init your view point here: 
	//viInitViewpoint(from, at, up, fov_angle, hither, resx, resy);

	return;

fmterr:
	printf("Parser view syntax error");
	exit(1);
}

/*----------------------------------------------------------------------
  parseLight()
  Positional light.  A light is defined by XYZ position and an optional color.
  For animated lights we need to be able to identify light sources. We do
  this by giving them a name.
  Description:
    "l" X Y Z [R G B]
    "la" name X Y Z [R G B]   #animated light
  Format:
    l %g %g %g [%g %g %g]
    la %s %g %g %g [%g %g %g]

    All light entities must be defined before any objects are defined (this
    requirement is so that NFF files can be used by hidden surface machines).
    Lights have a non-zero intensity of no particular value [this definition
    may change soon, with the addition of an intensity and/or color].
    The name of an animated light must not contain any white spaces.
----------------------------------------------------------------------*/
static void parseLight(FILE *fp) {
	float4 pos;
	float4 col;
	int num;
	int is_animated;
	char name[100];
	strcpy(name,"");

	is_animated = getc(fp);
	if(is_animated != 'a') {
		ungetc(is_animated, fp);
		is_animated=0;
	}

	if(is_animated) { // if it's an animated light, read its name
		fscanf(fp,"%s",name);
	}

	if(fscanf(fp, "%f %f %f ",&pos.x, &pos.y, &pos.z) != 3) {
		printf("Light source position syntax error");
		exit(1);
	}
	pos.w=1.0f;

	// read optional color of light
	num=fscanf(fp, "%f %f %f ",&col.x, &col.y, &col.z);
	if(num==0) {
		col = make_float4(1.0f,1.0f,1.0f,1.0f);
	} else if(num!=3) {
		printf("Light source color syntax error");
		exit(1);
	}
	col.z=1.0f;

	// add light source here:
//    viAddLight(name,pos,col);
}

static void eatWhitespace(FILE *f) {
	char ch=getc(f);
	while(ch==' ' || ch=='\t' || ch=='\n' || ch=='\f' || ch=='\r')
		ch=getc(f);
	ungetc(ch,f);
}

/*----------------------------------------------------------------------
  parseKeyFrames()
  Format:
  "k" xform_name
  {
     type num_keyframes
     keyframe_data
  }

  Description:  
  The xform_name identifies the xform which is to be animated.

  For each type of keyframes (transl, rot, scale) there must
  be at least 4 keyframes, and the first and the last keyframes
  are only used internally to get starting tangents and similar stuff.
  There is also a fourth type of keyframes (visibility), which determines
  whether an object is visible or not at a certain time. This one needs
  at least one keyframe. 
  Each of the types may appear once in a keyframe description, i.e.,
  you can have a translation, scaling, rotation, and a visibility
  in the keyframe description (but not, say, two rotations).
  It is always the case that the total transform
  is T*R*S, where T=translation, R=rotation, and S=scaling.
  Is there is no, say, rotation, then R=I (identity matrix). This
  holds for the other as well.
  Note the order: the scaling is applied first, then rotation,
  the translation.

  Translation example:
  k the_ball
  {
    transl 5
    -0.50 0 -3.0 0  0 0 0
     0.00 0  0.0 0  0 0 0
     0.50 0  1.0 0  0 0 0
     1.00 0  0.0 0  0 0 0
     1.50 0 -3.0 0  0 0 0
  }
  In this example, we can only use times from 0.00 to 1.00
  Each row looks like this: time x y z tension continuity bias,
  where (x,y,z) is the translation at time, and tension, continuity, and
  bias, are the constants for interpolation at time.
 
  Rotation example:
  k the_ball2
  {
     rot 4
     -0.5  1 0 0 45  0 0 0
      0.0  1 0 0 0   0 0 0
      0.5  1 0 0 90  0 0 0
      1.0  0 1 1 10  0 0 0
  }
  Each row looks like this: time x y z degrees tension continuity bias, where  
  (x,y,z) is the rotation axis and degrees is the amound which is rotated
  around the axis at time, and tension, continuity, and
  bias, are the constants for interpolation at time.

  Scaling example:
  k the_ball3
  {
    scale 7
    -0.5 1 1 1 0 0 0
     0.0 1 1 1 0 0 0
     0.5 2 1 1 0 0 0
     1.0 1 2 1 0 0 0
     1.5 1 1 2 0 0 0
     2.0 1 1 1 0 0 0
     2.5 1 1 1 0 0 0
  }
  Each row looks like this: time x y z tension continuity bias, where  
  (x,y,z) is the scaling parameters, and tension, continuity, and
  bias, are the constants for interpolation at time.

  Visibility example:
  k the_ball4
  {
    visibility 2
    0.5 0
    2.0 1
  }
  Each row looks like this: time visbility_flag
  where visibility_flag is either 0 (invisible) or 1 (visible)
  From time=-infinity each object is assumed visible until
  the first visibility keyframe. At that time (0.5 in the example
  above) the visibility switches to what is given in that keyframe
  (0 in the example). At the next keyframe (time=2.0 in the example)
  the visibility may change again (changes to visible (1) above).
  The last visibility_flag determines the visibility until time=infinity.

  Note also that if the name of an animation is "camera", then
  the viewpoint should be animated after those key frames (only
  translation and rotation). Light sources can also be animated
  (only translation).
----------------------------------------------------------------------*/

static void parseKeyFrames(FILE *fp) {   
	char name[200];
	char motion[200];
	char ch;
	int  c;
	int  ret, i, nKeyFrames;
	float time, x, y, z, angle, te, co, bi;
	PositionKey* PKeys;
	RotationKey* RKeys;
	Animation* animation;

	if(fscanf(fp,"%s",name)!=1) {
		printf("Error: could not read name of animation.\n");
		exit(1);
	}
	eatWhitespace(fp);
	ch=getc(fp);
	if(ch!='{') {
		printf("Error: could not find a { in animation %s.\n",name);
		exit(1);
	}
   
	// insert a new animation in the AnimationList
	animation = (Animation*) calloc(1, sizeof(Animation));
	animation->translations=NULL;
	animation->rotations=NULL;
	animation->scales=NULL;
	animation->name=(char *)malloc(sizeof(name)); 
	strcpy(animation->name,name);

	eatWhitespace(fp);
	while( (c = getc(fp)) != '}' ) {
		ungetc(c, fp);
		if(fscanf(fp, "%s %d", motion, &nKeyFrames)!=2) {
			printf("Error: could not read name of motion or number of keyframes for animation.\n");
			exit(1);
		}
      
		if(nKeyFrames<4 && strcmp("visibility",motion)) { 
			printf("Error: there must be at least 4 keyframes for %s.\n",name);
			exit(1);
		}

		// check whether the motion is a "transl" or a "rot" or a "scale"
		if(strcmp(motion, "transl")==0) {
			PKeys = (PositionKey*) calloc(nKeyFrames, sizeof(PositionKey));
			for( i=0; i<nKeyFrames; i++ ) {
				ret = fscanf(fp, " %f %f %f %f %f %f %f", &time, &x, &y, &z, &te, &co, &bi);
				if(ret != 7) {
					printf("error in parsing translation keyframes for %s\n", animation->name);
					exit(1);
				}

				PKeys[i].t = time;
				PKeys[i].P.x = x;
				PKeys[i].P.y = y;
				PKeys[i].P.z = z;
				PKeys[i].tension = te;
				PKeys[i].continuity = co;
				PKeys[i].bias = bi;
			}

			animation->translations = KB_PosInitialize(nKeyFrames, PKeys);
			free(PKeys);
		} else if(strcmp(motion, "rot")==0) {
			RKeys = (RotationKey*) calloc(nKeyFrames, sizeof(RotationKey));
			for( i=0; i<nKeyFrames; i++ ) {
				ret = fscanf(fp," %f %f %f %f %f %f %f %f", &time, &x, &y, &z, &angle, &te, &co, &bi);
				if(ret != 8) {
					printf("error in parsing rotation keyframes for %s\n",animation->name);
					exit(1);
				}

				RKeys[i].t = time;
				RKeys[i].Rot.x = x;
				RKeys[i].Rot.y = y;
				RKeys[i].Rot.z = z;
				RKeys[i].Rot.angle = angle*M_PI/180.0;
				RKeys[i].tension = te;
				RKeys[i].continuity = co;
				RKeys[i].bias = bi;
			}
			animation->rotations = KB_RotInitialize(nKeyFrames, RKeys);
			free(RKeys);
		} else if(strcmp(motion, "scale")==0) {
			PKeys = (PositionKey*) calloc(nKeyFrames, sizeof(PositionKey));
			for( i=0; i<nKeyFrames; i++ ) {
				ret = fscanf(fp, " %f %f %f %f %f %f %f", &time, &x, &y, &z, &te, &co, &bi);
				if(ret != 7) {
					printf("error in parsing scale keyframes for %s\n", animation->name);
					exit(1);
				}	
				PKeys[i].t = time;
				PKeys[i].P.x = x;
				PKeys[i].P.y = y;
				PKeys[i].P.z = z;
				PKeys[i].tension = te;
				PKeys[i].continuity = co;
				PKeys[i].bias = bi;
			}
			animation->scales = KB_PosInitialize(nKeyFrames, PKeys);
			free(PKeys);
		} else {
			printf("Error: unknown keyframe type (%s). Must be transl, rot, or scale.\n",motion);
			exit(1);
		}
		eatWhitespace(fp);
	}   
}

/*----------------------------------------------------------------------
  parseAnimParams()
  parse animation parameters
  Description:  
    "a" start_time end_time num_frames

Format:
    a %g %g %d 

    start_time indicates the start of the animation
    end_time   indicates the end of the animation
    num_frames is the number of frames in the animation.
      Note: the step time (from one frame to the next) is then
      (end_time-start_time)/(num_frames-1)
----------------------------------------------------------------------*/
static void parseAnimParams(FILE *fp) {
	float start,end;
	int num_frames;
	if(fscanf(fp,"%f %f %d",&start,&end,&num_frames)!=3) {
		printf("Error: could not parse animations parameters.\n");
		exit(1);
	}
	// add animations parameters here
//    viSetupAnimParams(start,end,num_frames);
}

/*----------------------------------------------------------------------
  parseA()
  parse animation parameters and global ambient light

  Global ambient light description:  
    "am" red green blue

Format:
    am %g %g %d 

    There is one global ambient light source in the scene,
    and it can be set with, e.g., "am 0.5 0.5 0.5".
    Default value is "am 1.0 1.0 1.0".
----------------------------------------------------------------------*/
static void parseA(FILE *f) {
	char name[100];
	char ch;
	int is_ambient;

	is_ambient = getc(f);
	if(is_ambient != 'm') {
		ungetc(is_ambient, f);
		is_ambient=0;
	}

	if(is_ambient) {  // we got "am" = global ambient light
		float3 amb;
		if(fscanf(f,"%f %f %f",&amb.x,&amb.y,&amb.z)!=3) {
			printf("Error: could not parse ambient light (am).\n");
			exit(1);
		}
		// set up your globabl ambient light here using amb
	} else {
		parseAnimParams(f);
	}
}

/*----------------------------------------------------------------------
  viParseFile()
  Description:
    parses the animation file
----------------------------------------------------------------------*/
bool viParseFile(FILE *f) {
   int ch;

   while((ch=getc(f))!=EOF) {  
		switch(ch) {
			case ' ':   // white space
			case '\t':
			case '\n':
			case '\f':
			case '\r':
				continue;
			case '#':   // comment
			case '%':   // comment
				parseComment(f); // ok
				break;
			case 'v':   // view point
				parseViewpoint(f); // ok
				break;
			case 'l':   // light source
				parseLight(f); // ok
				break;
			case 'a':  // animation parameters
				parseA(f);
				break;
			case 'k':  // key frames for transform (or the camera)
				parseKeyFrames(f);
				break;
			default:    // unknown
				printf("unknown NFF primitive code: %c\n",ch);
				exit(1);
		}
	}
	return true;
}
